# -*- coding: utf-8  -*-

import itertools
import os
import unittest
import logging
import random
import string
from unittest.mock import Mock, patch

import pytest

from xl.metadata import CoverImage
import xl.trax.track as track
import xl.settings as settings

LOG = logging.getLogger(__name__)


class Test_MetadataCacher:
    TIMEOUT = 2000
    MAX_ENTRIES = 2

    def setup_method(self):
        self.mc: track._MetadataCacher[str, str] = track._MetadataCacher(
            self.TIMEOUT, self.MAX_ENTRIES
        )

    def test_add(self):
        with patch('gi.repository.GLib.timeout_add_seconds') as timeout_add_seconds:
            self.mc.add('foo', 'bar')
            assert self.mc.get('foo') == 'bar'
        timeout_add_seconds.assert_called_once_with(
            self.TIMEOUT, self.mc._MetadataCacher__cleanup
        )

    def test_double_add(self):
        with patch('gi.repository.GLib.timeout_add_seconds') as timeout_add_seconds:
            self.mc.add('foo', 'bar')
            assert self.mc.get('foo') == 'bar'
            self.mc.add('foo', 'bar')
            assert self.mc.get('foo') == 'bar'
        timeout_add_seconds.assert_called_once_with(
            self.TIMEOUT, self.mc._MetadataCacher__cleanup
        )

    def test_overflow(self):
        """
        When the cache overflows, the least-recently used entry is removed
        """
        assert self.MAX_ENTRIES == 2
        with patch('gi.repository.GLib.timeout_add_seconds'), patch(
            'time.time', Mock(side_effect=itertools.count(1))
        ):
            self.mc.add('k1', 'v1')  # [entry1(time=1)]
            self.mc.add('k2', 'v2')  # [entry1(time=1), entry2(time=2)]
            assert self.mc.get('k2')  # [entry1(time=1), entry2(time=3)]
            assert self.mc.get('k1')  # [entry1(time=4), entry2(time=3)]
            # Adding another entry should remove entry2 because it has the
            # smallest time value
            self.mc.add('k3', 'v3')  # [entry1(time=4), entry3(time=5)]
            assert not self.mc.get('k2')
            assert self.mc.get('k1')
            assert self.mc.get('k3')

    def test_remove(self):
        with patch('gi.repository.GLib.timeout_add_seconds'), patch(
            'gi.repository.GLib.source_remove'
        ):
            self.mc.add('foo', 'bar')
            self.mc.remove('foo')
            assert self.mc.get('foo') is None

    def test_remove_not_exist(self):
        assert self.mc.remove('foo') is None


def random_str(l=8):
    return ''.join(random.choice(string.ascii_letters) for _ in range(l))


class TestTrack:
    def verify_tags_exist(self, tr, test_track, deleted=None):
        internal_tags = {'__length', '__modified', '__basedir', '__basename', '__loc'}
        if test_track.ext not in ['aac', 'spx']:
            internal_tags.add('__bitrate')

        disk_tags = {'album', 'tracknumber', 'artist', 'title'}
        normal_tags = internal_tags | disk_tags

        if test_track.has_cover:
            disk_tags.add('cover')

        if deleted:
            normal_tags.discard(deleted)
            disk_tags.remove(deleted)

        # check non-disk tag list
        assert set(tr.list_tags()) == normal_tags

        # check disk tag list
        assert set(tr.list_tags_disk()) == disk_tags

    ## Creation
    def test_flyweight(self, test_track):
        """There can only be one object based on a url in args"""
        t1 = track.Track(test_track.filename)
        t2 = track.Track(uri=test_track.uri)
        assert t1 is t2, "%s should be %s" % (repr(t1), repr(t2))

    def test_different_url_not_flyweighted(self, test_tracks):
        t1 = track.Track(test_tracks.get('.mp3').filename)
        t2 = track.Track(test_tracks.get('.ogg').filename)
        assert t1 is not t2, "%s should not be %s" % (repr(t1), repr(t2))

    def test_none_url(self):
        with pytest.raises(ValueError):
            track.Track()

    def test_pickles(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('artist', 'bar')
        assert tr._pickles() == {'__loc': 'file:///foo', 'artist': ['bar']}

    def test_unpickles(self):
        tr1 = track.Track(_unpickles={'artist': ['my_artist'], '__loc': 'uri'})
        assert tr1.get_loc_for_io() == 'uri'

    def test_unpickles_flyweight(self):
        tr1 = track.Track(_unpickles={'artist': ['my_artist'], '__loc': 'uri'})
        tr2 = track.Track(_unpickles={'artist': ['my_artist'], '__loc': 'uri'})
        assert tr1 is tr2

    def test_takes_nonurl(self, test_track):
        tr = track.Track(test_track.filename)

        assert tr.get_local_path()
        assert tr.exists()

    def test_takes_url(self, test_track):
        tr = track.Track(test_track.uri)

        assert tr.get_local_path()
        assert tr.exists()

    ## Information
    def test_local_type(self, test_track):
        tr = track.Track(test_track.filename)
        assert tr.get_type() == 'file'

    def test_is_local_local(self):
        """Tests a local filename -> True"""
        tr = track.Track('foo')
        assert tr.is_local() is True

    def test_is_local_remote(self):
        """Tests a remote filename -> False"""
        tr = track.Track('http://foo')
        assert tr.is_local() is False

    def test_local_filesize(self, test_track):
        tr = track.Track(test_track.filename)
        assert tr.get_size() == test_track.size

    def test_str(self, test_track):
        loc = test_track.filename
        tr = track.Track(loc)
        self.empty_track_of_tags(tr, ('__loc',))
        trstr = "<Track 'Unknown (%s)' by '' from ''>" % os.path.basename(loc)
        assert str(tr) == trstr
        tr.set_tag_raw('artist', 'art')
        tr.set_tag_raw('album', 'alb')
        tr.set_tag_raw('title', 'title')
        assert str(tr) == "<Track 'title' by 'art' from 'alb'>"

    def test_read_tags_no_perms(self, test_track_fp):
        if os.name != 'posix':
            pytest.skip("only works on POSIX")
        if os.geteuid() == 0:
            pytest.xfail("fails when run as root")

        tr = track.Track(test_track_fp.name)
        # first, ensure that we can actually read the tags to begin with
        assert tr.read_tags()

        os.chmod(test_track_fp.name, 0o000)

        # opening the file should fail...
        with pytest.raises(IOError):
            with open(test_track_fp.name, 'rb'):
                pass

        # second, ensure that we can no longer read them
        assert not tr.read_tags()

    def test_write_tags_no_perms(self, test_track_fp):
        if os.name != 'posix':
            pytest.skip("only works on POSIX")
        if os.geteuid() == 0:
            pytest.xfail("fails when run as root")

        os.chmod(test_track_fp.name, 0o444)

        tr = track.Track(test_track_fp.name)
        tr.set_tag_raw('artist', random_str())
        assert not tr.write_tags()

    def test_write_tag(self, writeable_track, writeable_track_name):
        artist = random_str()
        tr = track.Track(writeable_track_name)
        tr.set_tag_raw('artist', artist)
        assert tr.write_tags() is not False
        assert tr.get_tag_raw('artist') == [artist]

        # if tag was changed, ensure it gets overridden by reading tags from file
        tr.set_tag_raw('artist', None)
        assert tr.get_tag_raw('artist') is None

        assert tr.read_tags() is not False
        assert tr.get_tag_raw('artist') == [artist]

        self.verify_tags_exist(tr, writeable_track)

    def test_delete_tag(self, writeable_track, writeable_track_name):
        artist = random_str()
        tr = track.Track(writeable_track_name)
        assert tr.get_tag_raw('artist') is not None

        tr.set_tag_raw('artist', None)
        assert tr.write_tags() is not False
        assert tr.get_tag_raw('artist') is None

        # if tag was deleted, ensure it gets overridden by reading tags from file
        tr.set_tag_raw('artist', artist)
        assert tr.get_tag_raw('artist') == [artist]

        assert tr.read_tags() is not False
        assert tr.get_tag_raw('artist') is None

        self.verify_tags_exist(tr, writeable_track, deleted='artist')

    def test_delete_missing_tag(self, writeable_track, writeable_track_name):
        tr = track.Track(writeable_track_name)
        assert tr.get_tag_raw('genre') is None
        tr.set_tag_raw('genre', None)
        assert tr.get_tag_raw('genre') is None

        self.verify_tags_exist(tr, writeable_track)

        # this actually writes it to disk
        assert tr.write_tags() is not False

        # make sure another read works
        assert tr.read_tags() is not False

        self.verify_tags_exist(tr, writeable_track)

    def test_write_delete_cover(self, writeable_track, writeable_track_name):
        if not writeable_track.has_cover:
            return

        if writeable_track.ext in ['aac', 'mp4']:
            from mutagen.mp4 import MP4Cover

            newcover = CoverImage(
                None, None, 'image/jpeg', MP4Cover(random_str().encode('utf-8'))
            )
        else:
            newcover = CoverImage(
                3, 'cover', 'image/jpeg', random_str().encode('utf-8')
            )

        tr = track.Track(writeable_track_name)
        assert tr.get_tag_raw('cover') is None
        assert tr.get_tag_disk('cover') is not None

        # Can we delete a cover?
        tr.set_tag_disk('cover', None)
        assert tr.write_tags() is not False

        self.verify_tags_exist(tr, writeable_track, deleted='cover')

        # Can we write a new cover?
        tr.set_tag_disk('cover', newcover)
        assert tr.get_tag_raw('cover') is None
        assert tr.get_tag_disk('cover') == [newcover]

        # reading the tags shouldn't change anything, since we're reading from disk
        assert tr.read_tags() is not False
        assert tr.get_tag_raw('cover') is None
        assert tr.get_tag_disk('cover') == [newcover]

        self.verify_tags_exist(tr, writeable_track)

    def test_write_tag_invalid_format(self):
        tr = track.Track('/tmp/foo.foo')
        assert tr.write_tags() is False

    def test_join_tag_empty(self):
        """Tests get_tag_raw with join=True and an empty tag"""
        tr = track.Track('foo')
        assert tr.get_tag_raw('artist', join=True) is None

    def test_join_tag_one(self):
        """Tests get_tag_raw with join=True and one element in tag"""
        tr = track.Track('foo')
        tr.set_tag_raw('artist', 'foo')
        assert tr.get_tag_raw('artist', join=True) == 'foo'

    def test_join_tag_two(self):
        """Tests get_tag_raw with join=True and one element in tag"""
        tr = track.Track('foo')
        tr.set_tag_raw('artist', ['foo', 'bar'])
        assert tr.get_tag_raw('artist', join=True) == 'foo / bar'

    def empty_track_of_tags(self, track, exclude=None):
        """Removes all the tags from a track"""
        for tag in track.list_tags():
            if exclude is not None and tag in exclude:
                continue
            track.set_tag_raw(tag, None)

    def test_list_tags(self, test_track):
        tr = track.Track(test_track.filename)
        tags = {'artist': 'foo', 'album': 'bar', '__loc': test_track.filename}
        self.empty_track_of_tags(tr, tags)
        for tag, val in tags.items():
            tr.set_tag_raw(tag, val)
        assert set(tr.list_tags()) == {'album', '__loc', 'artist', '__basename'}

    def test_rating_empty(self):
        """Test get_rating when no rating has been set"""
        tr = track.Track('/foo')
        assert tr.get_rating() == 0

    def test_set_rating(self):
        tr = track.Track('/foo')
        tr.set_rating(2)
        assert tr.get_rating() == 2

    def test_set_rating_invalid(self):
        tr = track.Track('/bar')
        with pytest.raises(TypeError):
            tr.set_rating('foo')

    ## Tag Getting helper methods
    def test_split_numerical_none(self):
        assert track.Track.split_numerical(None) == (None, 0)

    def test_split_numerical_str(self):
        fn = track.Track.split_numerical
        assert fn('12/15') == (12, 15)
        assert fn('foo/15') == (None, 15)
        assert fn('12/foo') == (12, 0)
        assert fn('12/15/2009') == (12, 15)

    def test_split_numerical_list(self):
        fn = track.Track.split_numerical
        assert fn(['12/15']) == (12, 15)
        assert fn(['foo/15']) == (None, 15)
        assert fn(['12/foo']) == (12, 0)
        assert fn(['12/15/2009']) == (12, 15)

    def test_strip_leading(self):
        # Strips whitespace if it's an empty string
        value = " `~!@#$%^&*()_+-={}|[]\\\";'<>?,./"
        retvalue = "`~!@#$%^&*()_+-={}|[]\\\";'<>?,./"
        assert track.Track.strip_leading(value) == retvalue
        assert track.Track.strip_leading(value + "foo") == "foo"

    def test_cutter(self):
        value = 'the a foo'
        assert track.Track.the_cutter(value) == 'a foo'

    def test_expand_doubles(self):
        value = 'ßæĳŋœƕǆǉǌǳҥҵ'
        assert track.Track.expand_doubles(value) == 'ssaeijngoehvdzljnjdzngts'

    def test_lower(self):
        value = 'FooBar'
        assert track.Track.lower(value) == 'foobar FooBar'

    def test_cuts_cb(self):
        value = []
        settings.set_option('collection/strip_list', value)
        track.Track._the_cuts_cb(None, None, 'collection/strip_list')
        assert track.Track._Track__the_cuts == value

        value = ['the', 'foo']
        settings.set_option('collection/strip_list', value)
        track.Track._the_cuts_cb(None, None, 'collection/strip_list')
        assert track.Track._Track__the_cuts == value

    def test_strip_marks(self):
        value = 'The Hëllò Wóþλdâ'
        retval = 'The Hello Woþλda The Hëllò Wóþλdâ'
        assert track.Track.strip_marks(value) == retval

    ## Sort tags
    def test_get_sort_tag_no_join(self):
        tr = track.Track('/foo')
        value = 'hello'
        retval = ['hello hello hello hello']
        tr.set_tag_raw('artist', value)
        assert tr.get_tag_sort('artist', join=False) == retval

    def test_get_sort_tag_discnumber(self):
        tr = track.Track('/foo')
        value = '12/15'
        retval = 12
        tr.set_tag_raw('discnumber', value)
        assert tr.get_tag_sort('discnumber') == retval

    def test_get_sort_tag_tracknumber(self):
        tr = track.Track('/foo')

        value = '12/15'
        retval = 12
        tr.set_tag_raw('tracknumber', value)
        assert tr.get_tag_sort('tracknumber') == retval

    def test_get_sort_tag_artist(self):
        tr = track.Track('/foo')
        value = 'The Hëllò Wóþλdâ'
        retval = 'hello woþλda the hëllò wóþλdâ ' 'The Hello Woþλda The Hëllò Wóþλdâ'
        tr.set_tag_raw('artist', value)
        assert tr.get_tag_sort('artist') == retval

    def test_get_sort_tag_albumsort(self):
        tr = track.Track('/foo')
        value = 'the hello world'
        val_as = 'Foo Bar'
        retval = 'foo bar foo bar Foo Bar Foo Bar'
        tr.set_tag_raw('album', value)
        tr.set_tag_raw('albumsort', val_as)
        assert tr.get_tag_sort('album') == retval

    @unittest.skip("TODO")
    def test_get_sort_tag_compilation_unknown(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__compilation', 'foo')
        # Does not actually modify anything
        value = 'hello world'
        retval = ' '.join(['\uffff\uffff\uffff\ufffe'] * 4)
        tr.set_tag_raw('artist', value)
        assert tr.get_tag_sort('artist') == retval

    def test_get_sort_tag_compilation_known(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__compilation', 'foo')
        value = 'foo bar'
        retval = 'foo bar foo bar foo bar foo bar'
        tr.set_tag_raw('artist', 'hello world')
        tr.set_tag_raw('albumartist', 'albumartist')
        tr.set_tag_raw('artistsort', value)
        assert tr.get_tag_sort('artist') == retval

    def test_get_sort_tag_length(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__length', 36)
        assert tr.get_tag_sort('__length') == 36

    def test_get_sort_tag_playcount(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__playcount', 36)
        assert tr.get_tag_sort('__playcount') == 36

    def test_get_sort_tag_other(self):
        tr = track.Track('/foo')
        val = 'foobar'
        ret = 'foobar foobar foobar foobar'
        tr.set_tag_raw('coverart', val)
        assert tr.get_tag_sort('coverart') == ret

    ## Display Tags
    def test_get_display_tag_loc(self):
        import sys

        if sys.platform == 'win32':
            tr = track.Track('C:\\foo')
            assert tr.get_tag_display('__loc') == 'C:\\foo'
        else:
            tr = track.Track('/foo')
            assert tr.get_tag_display('__loc') == '/foo'
        tr = track.Track('http://foo/bar')
        assert tr.get_tag_display('__loc') == 'http://foo/bar'

    @unittest.skip("TODO")
    def test_get_display_tag_compilation(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__compilation', 'foo')
        assert tr.get_tag_display('artist') == track._VARIOUSARTISTSSTR

    def test_get_display_tag_discnumber(self):
        tr = track.Track('/foo')
        value = '12/15'
        retval = '12'
        tr.set_tag_raw('discnumber', value)
        assert tr.get_tag_display('discnumber') == retval

    def test_get_display_tag_tracknumber(self):
        tr = track.Track('/foo')
        value = '12/15'
        retval = '12'
        tr.set_tag_raw('tracknumber', value)
        assert tr.get_tag_display('tracknumber') == retval

    def test_get_display_tag_length(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__length', 360)
        assert tr.get_tag_display('__length') == '360'

    def test_get_display_tag_bitrate(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__bitrate', 48000)
        assert tr.get_tag_display('__bitrate') == '48k'

    def test_get_display_tag_bitrate_bitrateless_format(self, test_tracks):
        td = test_tracks.get('.flac')
        tr = track.Track(td.filename)
        assert tr.get_tag_display('__bitrate') == ''

    def test_get_display_tag_bitrate_bad(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__bitrate', 'lol')
        assert tr.get_tag_display('__bitrate') == ''

    def test_get_display_tag_numeric_zero(self):
        tr = track.Track('/foo')
        assert tr.get_tag_display('tracknumber') == ''
        assert tr.get_tag_display('discnumber') == ''
        assert tr.get_tag_display('__rating') == '0'
        assert tr.get_tag_display('__playcount') == '0'

    def test_get_display_tag_join_true(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('artist', ['foo', 'bar'])
        assert tr.get_tag_display('artist') == 'foo / bar'

    def test_get_display_tag_join_false(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('artist', ['foo', 'bar'])
        assert tr.get_tag_display('artist', join=False) == ['foo', 'bar']

    ## Sort tags
    def test_get_search_tag_loc(self):
        tr = track.Track('/foo')
        assert tr.get_tag_search('__loc') == '__loc=="file:///foo"'

    @unittest.skip("TODO")
    def test_get_search_tag_artist_compilation(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__compilation', 'foo')
        retval = 'albumartist=="albumartist" ! __compilation==__null__'
        tr.set_tag_raw('artist', 'hello world')
        tr.set_tag_raw('albumartist', 'albumartist')
        tr.set_tag_raw('artistsort', 'foo bar')
        assert tr.get_tag_search('artist') == retval

    def test_get_search_tag_artist(self):
        tr = track.Track('/foo')
        retval = 'artist=="hello world"'
        tr.set_tag_raw('artist', 'hello world')
        assert tr.get_tag_search('artist') == retval

    def test_get_search_tag_artist_none(self):
        tr = track.Track('/foo')
        retval = 'artist==__null__'
        assert tr.get_tag_search('artist') == retval

    def test_get_search_tag_discnumber(self):
        tr = track.Track('/foo')
        value = '12/15'
        retval = 'discnumber=="12"'
        tr.set_tag_raw('discnumber', value)
        assert tr.get_tag_search('discnumber') == retval

    def test_get_search_tag_tracknumber(self):
        tr = track.Track('/foo')
        value = '12/15'
        retval = 'tracknumber=="12"'
        tr.set_tag_raw('tracknumber', value)
        assert tr.get_tag_search('tracknumber') == retval

    def test_get_search_tag_length(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__length', 36)
        assert tr.get_tag_search('__length') == '__length=="36"'

    def test_get_search_tag_bitrate(self):
        tr = track.Track('/foo')
        tr.set_tag_raw('__bitrate', 48000)
        assert tr.get_tag_search('__bitrate') == '__bitrate=="48k" __bitrate=="48000"'

    def test_get_disk_tag_invalid_format(self):
        tr = track.Track('/tmp/foo.bah')
        assert tr.get_tag_disk('artist') is None

    def test_list_disk_tag_invalid_format(self):
        tr_name = '/tmp/foo.foo'
        tr = track.Track(tr_name)
        assert tr.list_tags_disk() is None

    def test_read_real_tracks(self, test_track):
        if not test_track.has_tags:
            return

        tr = track.Track(test_track.filename)

        # raw tags
        assert tr.get_tag_raw('album') == ['Chimera']
        assert tr.get_tag_raw('artist') == ['Delerium']
        assert tr.get_tag_raw('title') == ['Truly']
        assert tr.get_tag_raw('tracknumber') in [['5'], ['5/0']]

        # disk tags should be the same
        assert tr.get_tag_disk('album') == ['Chimera']
        assert tr.get_tag_disk('artist') == ['Delerium']
        assert tr.get_tag_disk('title') == ['Truly']
        assert tr.get_tag_disk('tracknumber') in [['5'], ['5/0']]

        self.verify_tags_exist(tr, test_track)



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

# Here's where it all begins.....
#
# Holds the main Exaile class, whose instantiation starts up the entirety
# of Exaile and which also handles Exaile shutdown.
#
# Also takes care of parsing commandline options.

import os
import platform
import sys
import threading

from .version import register

from xl import logger_setup
from xl.externals.sigint import InterruptibleLoopContext
from xl.nls import gettext as _

# Imported later to avoid PyGObject imports just for --help.
GLib = Gio = Gtk = common = xdg = None


def _do_heavy_imports():
    global GLib, Gio, Gtk, common, xdg

    import gi

    gi.require_version('Gdk', '3.0')
    gi.require_version('Gtk', '3.0')
    gi.require_version('Gst', '1.0')
    gi.require_version('GstPbutils', '1.0')

    from gi.repository import GLib, Gio, Gtk
    from xl import common, xdg


# placeholder, - xl.version can be slow to import, which would slow down
# cli args. Thus we import __version__ later.
__version__ = None

logger = None


def create_argument_parser():
    """Create command-line argument parser for Exaile"""

    import argparse

    # argparse hard-codes "usage:" uncapitalized. We replace this with an
    # empty string and put "Usage:" in the actual usage string instead.

    class Formatter(argparse.HelpFormatter):
        def _format_usage(self, usage, actions, groups, prefix):
            return super(self.__class__, self)._format_usage(usage, actions, groups, "")

    p = argparse.ArgumentParser(
        usage=_("Usage: exaile [OPTION...] [LOCATION...]"),
        description=_(
            "Launch Exaile, optionally adding tracks specified by"
            " LOCATION to the active playlist."
            " If Exaile is already running, this attempts to use the existing"
            " instance instead of creating a new one."
        ),
        add_help=False,
        formatter_class=Formatter,
    )

    p.add_argument('locs', nargs='*', help=argparse.SUPPRESS)

    group = p.add_argument_group(_('Playback Options'))
    group.add_argument(
        "-n",
        "--next",
        dest="Next",
        action="store_true",
        default=False,
        help=_("Play the next track"),
    )
    group.add_argument(
        "-p",
        "--prev",
        dest="Prev",
        action="store_true",
        default=False,
        help=_("Play the previous track"),
    )
    group.add_argument(
        "-s",
        "--stop",
        dest="Stop",
        action="store_true",
        default=False,
        help=_("Stop playback"),
    )
    group.add_argument(
        "-a", "--play", dest="Play", action="store_true", default=False, help=_("Play")
    )
    group.add_argument(
        "-u",
        "--pause",
        dest="Pause",
        action="store_true",
        default=False,
        help=_("Pause"),
    )
    group.add_argument(
        "-t",
        "--play-pause",
        dest="PlayPause",
        action="store_true",
        default=False,
        help=_("Pause or resume playback"),
    )
    group.add_argument(
        "--stop-after-current",
        dest="StopAfterCurrent",
        action="store_true",
        default=False,
        help=_("Stop playback after current track"),
    )

    group = p.add_argument_group(_('Collection Options'))
    group.add_argument(
        "--add",
        dest="Add",
        # TRANSLATORS: Meta variable for --add and --export-playlist
        metavar=_("LOCATION"),
        help=_("Add tracks from LOCATION to the collection"),
    )

    group = p.add_argument_group(_('Playlist Options'))
    group.add_argument(
        "--export-playlist",
        dest="ExportPlaylist",
        # TRANSLATORS: Meta variable for --add and --export-playlist
        metavar=_("LOCATION"),
        help=_('Export the current playlist to LOCATION'),
    )

    group = p.add_argument_group(_('Track Options'))
    group.add_argument(
        "-q",
        "--query",
        dest="Query",
        action="store_true",
        default=False,
        help=_("Query player"),
    )
    group.add_argument(
        "--format-query",
        dest="FormatQuery",
        # TRANSLATORS: Meta variable for --format-query
        metavar=_('FORMAT'),
        help=_('Retrieve the current playback state and track information as FORMAT'),
    )
    group.add_argument(
        "--format-query-tags",
        dest="FormatQueryTags",
        # TRANSLATORS: Meta variable for --format-query-tags
        metavar=_('TAGS'),
        help=_('Tags to retrieve from the current track; use with --format-query'),
    )
    group.add_argument(
        "--gui-query",
        dest="GuiQuery",
        action="store_true",
        default=False,
        help=_("Show a popup with data of the current track"),
    )
    group.add_argument(
        "--get-title",
        dest="GetTitle",
        action="store_true",
        default=False,
        help=_("Print the title of current track"),
    )
    group.add_argument(
        "--get-album",
        dest="GetAlbum",
        action="store_true",
        default=False,
        help=_("Print the album of current track"),
    )
    group.add_argument(
        "--get-artist",
        dest="GetArtist",
        action="store_true",
        default=False,
        help=_("Print the artist of current track"),
    )
    group.add_argument(
        "--get-length",
        dest="GetLength",
        action="store_true",
        default=False,
        help=_("Print the length of current track"),
    )
    group.add_argument(
        '--set-rating',
        dest="SetRating",
        type=int,
        # TRANSLATORS: Variable for command line options with arguments
        metavar=_('N'),
        help=_('Set rating for current track to N%').replace("%", "%%"),
    )
    group.add_argument(
        '--get-rating',
        dest='GetRating',
        action='store_true',
        default=False,
        help=_('Get rating for current track'),
    )
    group.add_argument(
        "--current-position",
        dest="CurrentPosition",
        action="store_true",
        default=False,
        help=_("Print the current playback position as time"),
    )
    group.add_argument(
        "--current-progress",
        dest="CurrentProgress",
        action="store_true",
        default=False,
        help=_("Print the current playback progress as percentage"),
    )

    group = p.add_argument_group(_('Volume Options'))
    group.add_argument(
        "-i",
        "--increase-vol",
        dest="IncreaseVolume",
        type=int,
        # TRANSLATORS: Meta variable for --increase-vol and--decrease-vol
        metavar=_("N"),
        help=_("Increase the volume by N%").replace("%", "%%"),
    )
    group.add_argument(
        "-l",
        "--decrease-vol",
        dest="DecreaseVolume",
        type=int,
        # TRANSLATORS: Meta variable for --increase-vol and--decrease-vol
        metavar=_("N"),
        help=_("Decrease the volume by N%").replace("%", "%%"),
    )
    group.add_argument(
        "-m",
        "--toggle-mute",
        dest="ToggleMute",
        action="store_true",
        default=False,
        help=_("Mute or unmute the volume"),
    )
    group.add_argument(
        "--get-volume",
        dest="GetVolume",
        action="store_true",
        default=False,
        help=_("Print the current volume percentage"),
    )

    group = p.add_argument_group(_('Other Options'))
    group.add_argument(
        "--new",
        dest="NewInstance",
        action="store_true",
        default=False,
        help=_("Start new instance"),
    )
    group.add_argument(
        "-h", "--help", action="help", help=_("Show this help message and exit")
    )
    group.add_argument(
        "--version",
        dest="ShowVersion",
        action="store_true",
        help=_("Show program's version number and exit."),
    )
    group.add_argument(
        "--start-minimized",
        dest="StartMinimized",
        action="store_true",
        default=False,
        help=_("Start minimized (to tray, if possible)"),
    )
    group.add_argument(
        "--toggle-visible",
        dest="GuiToggleVisible",
        action="store_true",
        default=False,
        help=_("Toggle visibility of the GUI (if possible)"),
    )
    group.add_argument(
        "--safemode",
        dest="SafeMode",
        action="store_true",
        default=False,
        help=_(
            "Start in safe mode - sometimes" " useful when you're running into problems"
        ),
    )
    group.add_argument(
        "--force-import",
        dest="ForceImport",
        action="store_true",
        default=False,
        help=_(
            "Force import of old data" " from version 0.2.x (overwrites current data)"
        ),
    )
    group.add_argument(
        "--no-import",
        dest="NoImport",
        action="store_true",
        default=False,
        help=_("Do not import old data" " from version 0.2.x"),
    )
    group.add_argument(
        "--start-anyway",
        dest="StartAnyway",
        action="store_true",
        default=False,
        help=_("Make control options like" " --play start Exaile if it is not running"),
    )
    group.add_argument(
        "--use-lang",
        dest="UseLanguage",
        default='',
        help=_("Locale to use for Exaile"),
    )

    group = p.add_argument_group(_('Development/Debug Options'))
    group.add_argument(
        "--datadir",
        dest="UseDataDir",
        metavar=_('DIRECTORY'),
        help=_("Set data directory"),
    )
    group.add_argument(
        "--all-data-dir",
        dest="UseAllDataDir",
        metavar=_('DIRECTORY'),
        help=_("Set data and config directory"),
    )
    group.add_argument(
        "--modulefilter",
        dest="ModuleFilter",
        metavar=_('MODULE'),
        help=_('Limit log output to MODULE'),
    )
    group.add_argument(
        "--levelfilter",
        dest="LevelFilter",
        metavar=_('LEVEL'),
        help=_('Limit log output to LEVEL'),
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    )
    group.add_argument(
        "--debug",
        dest="Debug",
        action="store_true",
        default=False,
        help=_("Show debugging output"),
    )
    group.add_argument(
        "--eventdebug",
        dest="DebugEvent",
        action="store_true",
        default=False,
        help=_("Enable debugging of" " xl.event. Generates lots of output"),
    )
    group.add_argument(
        "--eventdebug-full",
        dest="DebugEventFull",
        action="store_true",
        default=False,
        help=_("Enable full debugging of" " xl.event. Generates LOTS of output"),
    )
    group.add_argument(
        "--threaddebug",
        dest="DebugThreads",
        action="store_true",
        default=False,
        help=_("Add thread name to logging" " messages."),
    )
    group.add_argument(
        "--eventfilter",
        dest="EventFilter",
        metavar=_('TYPE'),
        help=_("Limit xl.event debug to output of TYPE"),
    )
    group.add_argument(
        "--quiet",
        dest="Quiet",
        action="store_true",
        default=False,
        help=_("Reduce level of output"),
    )
    group.add_argument(
        '--startgui', dest='StartGui', action='store_true', default=False
    )
    group.add_argument(
        '--no-dbus',
        dest='Dbus',
        action='store_false',
        default=True,
        help=_("Disable D-Bus support"),
    )
    group.add_argument(
        '--no-hal',
        dest='Hal',
        action='store_false',
        default=True,
        help=_("Disable HAL support."),
    )

    return p


class Exaile:
    _exaile = None

    def __init__(self):
        """
        Initializes Exaile.
        """
        self.quitting = False
        self.loading = True

        # NOTE: This automatically exits on --help.
        self.options = create_argument_parser().parse_args()

        if self.options.ShowVersion:
            self.version()
            return

        _do_heavy_imports()

        # Set program name for matching with .desktop file in Plasma
        # under wayland (see #653); should be done before splash screen
        # is displayed.
        GLib.set_prgname('exaile')

        if self.options.UseDataDir:
            xdg.data_dirs.insert(1, self.options.UseDataDir)

        # this is useful on Win32, because you cannot set these directories
        # via environment variables
        if self.options.UseAllDataDir:
            alldatadir = self.options.UseAllDataDir

            # TODO: is this still necessary? Python3 does not seem to
            # have issue with UTF-8 characters in path (in contrast
            # to Python2, os.path.join() does not fail).
            # For now, we replace the UTF-8 characters with ? to keep
            # the behavior consistent with the old version...
            if not os.path.supports_unicode_filenames:
                try:
                    alldatadir.encode('ascii')
                except UnicodeEncodeError:
                    # Replace non-ASCII characters with ?
                    alldatadir = alldatadir.encode('ascii', 'replace').decode('ascii')
                    print(
                        "WARNING : converted non-ASCII data dir %s to ascii: %s"
                        % (self.options.UseAllDataDir, alldatadir)
                    )
            xdg.data_home = alldatadir
            xdg.data_dirs.insert(0, xdg.data_home)
            xdg.config_home = alldatadir
            xdg.config_dirs.insert(0, xdg.config_home)
            xdg.cache_home = alldatadir
            xdg.plugin_dirs[0] = os.path.join(alldatadir, 'plugins')

        try:
            xdg._make_missing_dirs()
        except OSError as e:
            print(
                'ERROR: Could not create configuration directories: %s' % e,
                file=sys.stderr,
            )
            return

        # Make event debug imply debug
        if self.options.DebugEventFull:
            self.options.DebugEvent = True

        if self.options.DebugEvent:
            self.options.Debug = True

        try:
            logger_setup.start_logging(
                self.options.Debug,
                self.options.Quiet,
                self.options.DebugThreads,
                self.options.ModuleFilter,
                self.options.LevelFilter,
            )
        except OSError as e:
            print('ERROR: could not setup logging: %s' % e, file=sys.stderr)
            return

        global logger
        import logging

        logger = logging.getLogger(__name__)

        try:
            # Late import ensures xl.event uses correct logger
            from xl import event

            if self.options.EventFilter:
                event.EVENT_MANAGER.logger_filter = self.options.EventFilter
                self.options.DebugEvent = True

            if self.options.DebugEvent:
                event.EVENT_MANAGER.use_logger = True

            if self.options.DebugEventFull:
                event.EVENT_MANAGER.use_verbose_logger = True

            # initial mainloop setup. The actual loop is started later,
            # if necessary
            self.mainloop_init()

            # initialize DbusManager
            if self.options.StartGui and self.options.Dbus:
                from xl import xldbus

                exit = xldbus.check_exit(self.options, self.options.locs)
                if exit == "exit":
                    sys.exit(0)
                elif exit == "command":
                    if not self.options.StartAnyway:
                        sys.exit(0)
                self.dbus = xldbus.DbusManager(self)

            # import version, see note above
            global __version__
            from xl.version import __version__

            # load the rest.
            self.__init()

            # handle delayed commands
            if (
                self.options.StartGui
                and self.options.Dbus
                and self.options.StartAnyway
                and exit == "command"
            ):
                xldbus.run_commands(self.options, self.dbus)

            # connect dbus signals
            if self.options.StartGui and self.options.Dbus:
                self.dbus._connect_signals()

            # On SIGTERM, quit normally.
            import signal

            signal.signal(signal.SIGTERM, (lambda sig, stack: self.quit()))

            # run the GUIs mainloop, if needed
            if self.options.StartGui:
                # Handle keyboard interruptions
                with InterruptibleLoopContext(self.quit):
                    Gtk.main()  # mainloop
        except Exception:
            logger.exception("Unhandled exception")

    def __init(self):
        """
        Initializes Exaile
        """

        logger.info("Loading Exaile %s...", __version__)

        from gi.repository import GObject

        register('Python', platform.python_version())
        register('PyGObject', '%d.%d.%d' % GObject.pygobject_version)

        logger.info("Loading settings...")
        try:
            from xl import settings
        except common.VersionError:
            logger.exception("Error loading settings")
            sys.exit(1)

        logger.debug("Settings loaded from %s", settings.location)

        self._set_locale(self.options.UseLanguage)

        splash = None

        if self.options.StartGui:
            if settings.get_option('gui/use_splash', True):
                from xlgui.widgets.info import Splash

                splash = Splash()
                splash.show()

        firstrun = settings.get_option("general/first_run", True)

        # Migrate old rating options
        from xl.migrations.settings import rating

        rating.migrate()

        # Migrate builtin OSD to plugin
        from xl.migrations.settings import osd

        osd.migrate()

        # Migrate engines
        from xl.migrations.settings import engine

        engine.migrate()

        # Activate new flac/ogg metadata handling for new installations
        from xl.migrations.settings import flac_tempo

        flac_tempo.migrate()

        # TODO: enable audio plugins separately from normal
        #       plugins? What about plugins that use the player?

        # Gstreamer doesn't initialize itself automatically, and fails
        # miserably when you try to inherit from something and GST hasn't
        # been initialized yet. So this is here.
        from gi.repository import Gst

        Gst.init(None)

        # Initialize plugin manager
        from xl import plugins

        self.plugins = plugins.PluginsManager(self)

        if not self.options.SafeMode:
            logger.info("Loading plugins...")
            self.plugins.load_enabled()
        else:
            logger.info("Safe mode enabled, not loading plugins.")

        # Initialize the collection
        logger.info("Loading collection...")
        from xl import collection

        try:
            self.collection = collection.Collection(
                "Collection", location=os.path.join(xdg.get_data_dir(), 'music.db')
            )
        except common.VersionError:
            logger.exception("VersionError loading collection")
            sys.exit(1)

        # Migrate covers.db. This can only be done after the collection is loaded.
        import xl.migrations.database.covers_1to2 as mig

        mig.migrate()

        from xl import event

        # Set up the player and playback queue
        from xl import player

        event.log_event("player_loaded", player.PLAYER, None)

        # Initialize playlist manager
        from xl import playlist

        self.playlists = playlist.PlaylistManager()
        self.smart_playlists = playlist.SmartPlaylistManager(
            'smart_playlists', collection=self.collection
        )
        if firstrun:
            self._add_default_playlists()
        event.log_event("playlists_loaded", self, None)

        # Initialize dynamic playlist support
        from xl import dynamic

        dynamic.MANAGER.collection = self.collection

        # Initialize device manager
        logger.info("Loading devices...")
        from xl import devices

        self.devices = devices.DeviceManager()
        event.log_event("device_manager_ready", self, None)

        # Initialize dynamic device discovery interface
        # -> if initialized and connected, then the object is not None

        self.udisks2 = None

        if self.options.Hal:
            from xl import hal

            udisks2 = hal.UDisks2(self.devices)
            if udisks2.connect():
                self.udisks2 = udisks2

        # Radio Manager
        from xl import radio

        self.stations = playlist.PlaylistManager('radio_stations')
        self.radio = radio.RadioManager()

        self.gui = None
        # Setup GUI
        if self.options.StartGui:
            logger.info("Loading interface...")

            import xlgui

            self.gui = xlgui.Main(self)
            if not self.options.StartMinimized:
                self.gui.main.window.show_all()

                if not settings.get_option('general/hide_413_dialog', False):
                    from xlgui.widgets import dialogs

                    dialogs.UpdateInfoDialog(self.gui.main.window)

            event.log_event("gui_loaded", self, None)

            if splash is not None:
                splash.destroy()

        if firstrun:
            settings.set_option("general/first_run", False)

        self.loading = False
        Exaile._exaile = self
        event.log_event("exaile_loaded", self, None)

        restore = True

        if self.gui:
            # Find out if the user just passed in a list of songs
            # TODO: find a better place to put this

            songs = [Gio.File.new_for_path(arg).get_uri() for arg in self.options.locs]
            if len(songs) > 0:
                restore = False
                self.gui.open_uri(songs[0], play=True)
                for arg in songs[1:]:
                    self.gui.open_uri(arg)

            # kick off autoscan of libraries
            # -> don't do it in command line mode, since that isn't expected
            self.gui.rescan_collection_with_progress(True)

        restore_play_state = settings.get_option("player/resume_playback", True)

        if restore and restore_play_state:
            player.QUEUE._restore_player_state(
                os.path.join(xdg.get_data_dir(), 'player.state')
            )

            if self.gui:
                GLib.idle_add(self.gui.get_playlist_container().show_current_track)

        # pylint: enable-msg=W0201

    def _set_locale(self, custom_lang: str = None) -> None:
        """
        Get and set locale setting
        """
        try:
            import locale

            lc, enc = locale.getlocale()

            if custom_lang:
                lc = custom_lang

            if enc is not None:
                locale_str = '%s %s' % (lc, enc)
            else:
                locale_str = _('Unknown')

            os.environ['LANGUAGE'] = lc
            register('Locale', locale_str)
        except Exception:
            pass

    def version(self):
        from xl.version import __version__

        print("Exaile", __version__)
        sys.exit(0)

    def _add_default_playlists(self):
        """
        Adds some default smart playlists to the playlist manager
        """
        from xl import playlist

        # entire playlist
        entire_lib = playlist.SmartPlaylist(
            _("Entire Library"), collection=self.collection
        )
        self.smart_playlists.save_playlist(entire_lib, overwrite=True)

        # random playlists
        for count in (100, 300, 500):
            pl = playlist.SmartPlaylist(
                _("Random %d") % count, collection=self.collection
            )
            pl.set_return_limit(count)
            pl.set_random_sort(True)
            self.smart_playlists.save_playlist(pl, overwrite=True)

        # rating based playlists
        for item in (3, 4):
            pl = playlist.SmartPlaylist(
                _("Rating > %d") % item, collection=self.collection
            )
            pl.add_param('__rating', '>', item)
            self.smart_playlists.save_playlist(pl, overwrite=True)

    def mainloop_init(self):
        from gi.repository import GObject

        MIN_VER = (3, 10, 2)
        ver = GObject.pygobject_version

        if ver < MIN_VER:
            # Probably should exit?
            logger.warning(
                "Exaile requires PyGObject %d.%d.%d or greater! (got %d.%d.%d)",
                *(MIN_VER + ver)
            )

        if self.options.Dbus:
            import dbus
            import dbus.mainloop.glib

            dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
            dbus.mainloop.glib.threads_init()
            dbus.mainloop.glib.gthreads_init()

        if not self.options.StartGui:
            from gi.repository import GLib

            loop = GLib.MainLoop()
            context = loop.get_context()
            t = threading.Thread(target=self.__mainloop, args=(context,))
            t.daemon = True
            t.start()

    def __mainloop(self, context):
        while True:
            try:
                context.iteration(True)
            except Exception:
                pass

    def get_version(self):
        """
        Returns the current version
        """
        return __version__

    def get_user_agent_string(self, plugin_name=None):
        """
        Returns an appropriately formatted User-agent string for
        web requests. When possible, plugins should use this to
        format user agent strings.

        Users can control this agent string by manually setting
        general/user_agent and general/user_agent_w_plugin in settings.ini

        :param plugin_name: the name of the plugin
        """

        version = __version__
        if '+' in version:  # strip out revision identifier
            version = version[: version.index('+')]

        fmt = {'version': version}

        if not hasattr(self, '_user_agent_no_plugin'):
            from xl import settings

            default_no_plugin = 'Exaile/%(version)s (+https://www.exaile.org)'
            default_plugin = 'Exaile/%(version)s %(plugin_name)s/%(plugin_version)s (+https://www.exaile.org)'

            self._user_agent_no_plugin = settings.get_option(
                'general/user_agent', default_no_plugin
            )
            self._user_agent_w_plugin = settings.get_option(
                'general/user_agent_w_plugin', default_plugin
            )

        if plugin_name is not None:
            plugin_info = self.plugins.get_plugin_info(plugin_name)

            fmt['plugin_name'] = plugin_info['Name'].replace(' ', '')
            fmt['plugin_version'] = plugin_info['Version']

            return self._user_agent_w_plugin % fmt
        else:
            return self._user_agent_no_plugin % fmt

    def quit(self, restart=False):
        """
        Exits Exaile normally. Takes care of saving
        preferences, databases, etc.

        :param restart: Whether to directly restart
        :type restart: bool
        """
        if self.quitting:
            return
        self.quitting = True
        logger.info("Exaile is shutting down...")

        logger.info("Tearing down plugins...")
        self.plugins.teardown(self)

        from xl import event

        # this event should be used by modules that dont need
        # to be saved in any particular order. modules that might be
        # touched by events triggered here should be added statically
        # below.
        event.log_event("quit_application", self, None)

        logger.info("Saving state...")
        self.plugins.save_enabled()

        if self.gui:
            self.gui.quit()

        from xl import covers

        covers.MANAGER.save()

        self.collection.save_to_location()

        # Save order of custom playlists
        self.playlists.save_order()
        self.stations.save_order()

        # save player, queue
        from xl import player

        player.QUEUE._save_player_state(
            os.path.join(xdg.get_data_dir(), 'player.state')
        )
        player.QUEUE.save_to_location(os.path.join(xdg.get_data_dir(), 'queue.state'))
        player.PLAYER.stop()

        from xl import settings

        settings.MANAGER.save()

        if restart:
            logger.info("Restarting...")
            logger_setup.stop_logging()
            python = sys.executable
            if sys.platform == 'win32':
                # Python Win32 bug: it does not quote individual command line
                # arguments. Here we do it ourselves and pass the whole thing
                # as one string.
                # See https://bugs.python.org/issue436259 (closed wontfix).
                import subprocess

                cmd = [python] + sys.argv
                cmd = subprocess.list2cmdline(cmd)
                os.execl(python, cmd)
            else:
                os.execl(python, python, *sys.argv)

        logger.info("Bye!")
        logger_setup.stop_logging()
        sys.exit(0)


def exaile():
    if not Exaile._exaile:
        raise AttributeError(
            "Exaile is not yet finished loading. "
            "Perhaps you should listen for the exaile_loaded signal?"
        )

    return Exaile._exaile


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Provides the fundamental objects for handling a list of tracks contained
in playlists as well as methods to import and export from various file formats.
"""

from gi.repository import Gio

from collections import deque
from datetime import datetime, timedelta
import logging
import operator
import os
import pickle
import random
import re
import sys
import time
from typing import NamedTuple
import urllib.parse
import urllib.request

from xl import common, dynamic, event, main, providers, settings, trax, xdg
from xl.common import GioFileInputStream, GioFileOutputStream, MetadataList
from xl.nls import gettext as _
from xl.metadata.tags import tag_data

logger = logging.getLogger(__name__)


class InvalidPlaylistTypeError(Exception):
    pass


class PlaylistExists(Exception):
    pass


class UnknownPlaylistTrackError(Exception):
    pass


class PlaylistExportOptions(NamedTuple):
    relative: bool


def encode_filename(filename: str) -> str:
    """
    Converts a file name into a valid filename most
    likely to not cause problems on any platform.

    :param filename: the name of the file
    """
    # list of invalid chars that need to be encoded
    # Note: '%' is the prefix for encoded chars so blacklist it too
    blacklist = r'<>:"/\|?*%'

    def encode_char(c):
        return '%' + hex(ord(c))[2:] if c in blacklist else c

    # encode any blacklisted chars
    filename = ''.join(map(encode_char, filename)) + '.playlist'

    return filename


def is_valid_playlist(path: str) -> bool:
    """
    Returns whether the file at a given path is a valid
    playlist. Checks for content type and falls back to
    file extension if unknown.

    :param path: the source path
    """
    content_type = Gio.content_type_guess(path)[0]

    if not Gio.content_type_is_unknown(content_type):
        for provider in providers.get('playlist-format-converter'):
            if content_type in provider.content_types:
                return True

    file_extension = path.split('.')[-1]

    for provider in providers.get('playlist-format-converter'):
        if file_extension in provider.file_extensions:
            return True

    return False


def import_playlist(path: str) -> 'Playlist':
    """
    Determines the type of playlist and creates
    a playlist from it

    :param path: the source path
    :returns: the playlist
    """
    # First try the cheap Gio way
    content_type = Gio.content_type_guess(path)[0]

    if not Gio.content_type_is_unknown(content_type):
        for provider in providers.get('playlist-format-converter'):
            if content_type in provider.content_types:
                return provider.import_from_file(path)

    # Next try to extract the file extension via URL parsing
    file_extension = urllib.parse.urlparse(path).path.split('.')[-1]

    for provider in providers.get('playlist-format-converter'):
        if file_extension in provider.file_extensions:
            return provider.import_from_file(path)

    # Last try the expensive Gio way (downloads the data for inspection)
    content_type = (
        Gio.File.new_for_uri(path)
        .query_info('standard::content-type')
        .get_content_type()
    )

    if content_type:
        for provider in providers.get('playlist-format-converter'):
            if content_type in provider.content_types:
                return provider.import_from_file(path)

    raise InvalidPlaylistTypeError(_('Invalid playlist type.'))


def export_playlist(playlist, path: str, options=None) -> None:
    """
    Exact same as @see import_playlist except
    it exports
    """
    file_extension = path.split('.')[-1]

    if hasattr(playlist, 'get_playlist'):
        playlist = playlist.get_playlist()
        if playlist is None:
            raise InvalidPlaylistTypeError(
                "SmartPlaylist not associated with a collection"
            )

    for provider in providers.get('playlist-format-converter'):
        if file_extension in provider.file_extensions:
            provider.export_to_file(playlist, path, options)
            break
    else:
        raise InvalidPlaylistTypeError(_('Invalid playlist type.'))


class FormatConverter:
    """
    Base class for all converters allowing to
    import from and export to a specific format
    """

    title = _('Playlist')
    content_types = []
    file_extensions = property(lambda self: [self.name])

    def __init__(self, name):
        self.name = name

    def export_to_file(self, playlist, path, options=None):
        """
        Export a playlist to a given path

        :param playlist: the playlist
        :type playlist: :class:`Playlist`
        :param path: the target path
        :type path: string
        :param options: exporting options
        :type options: :class:`PlaylistExportOptions`
        """
        pass

    def import_from_file(self, path):
        """
        Import a playlist from a given path

        :param path: the source path
        :type path: string
        :returns: the playlist
        :rtype: :class:`Playlist`
        """
        pass

    def name_from_path(self, path):
        """
        Convenience method to retrieve a sane
        name from a path

        :param path: the source path
        :type path: string
        :returns: a name
        :rtype: string
        """
        gfile = Gio.File.new_for_uri(path)
        name = gfile.get_basename()

        for extension in self.file_extensions:
            if name.endswith(extension):
                # Remove known extension
                return name[: -len(extension) - 1]
        return name

    def get_track_import_path(self, playlist_path, track_path):
        """
        Retrieves the import path of a track

        :param playlist_path: the import path of the playlist
        :type playlist_path: string
        :param track_path: the path of the track
        :type track_path: string
        """
        playlist_uri = Gio.File.new_for_uri(playlist_path).get_uri()
        # Track path will not be changed if it already is a fully qualified URL
        track_uri = urllib.parse.urljoin(playlist_uri, track_path.replace('\\', '/'))

        logging.debug('Importing track: %s' % track_uri)

        # Now, let's be smart about importing the file/playlist. If the
        # original URI cannot be found and its a local path, then do a
        # small search for the track relative to the playlist to see
        # if it can be found.

        # TODO: Scan collection for tracks as last resort??

        if track_uri.startswith('file:///') and not Gio.File.new_for_uri(
            track_uri
        ).query_exists(None):
            if not playlist_uri.startswith('file:///'):
                logging.debug('Track does not seem to exist, using original path')
            else:
                logging.debug(
                    'Track does not seem to exist, trying different path combinations'
                )

                def _iter_uris(pp, tp):
                    pps = pp[len('file:///') :].split('/')
                    tps = tp.strip().replace('\\', '/').split('/')

                    # handle absolute paths correctly
                    if tps[0] == '':
                        tps = tps[1:]

                    # iterate the playlist path a/b/c/d, a/b/c, a/b, ...
                    for p in range(len(pps) - 1, 0, -1):
                        ppp = 'file:///%s' % '/'.join(pps[0:p])

                        # iterate the file path d, c/d, b/c/d, ...
                        for t in range(len(tps) - 1, -1, -1):
                            yield '%s/%s' % (ppp, '/'.join(tps[t : len(tps)]))

                for uri in _iter_uris(playlist_uri, track_path):
                    logging.debug('Trying %s' % uri)
                    if Gio.File.new_for_uri(uri).query_exists(None):
                        track_uri = uri
                        logging.debug('Track found at %s' % uri)
                        break

        return track_uri

    def get_track_export_path(
        self, playlist_uri: str, track_uri: str, options: PlaylistExportOptions
    ):
        """
        Retrieves the export path of a track,
        possibly influenced by options

        :param playlist_uri: the export path of the playlist
        :param track_uri: the path of the track
        :param options: options
        """

        track_uri_components = urllib.parse.urlparse(track_uri)
        export_uri_components = urllib.parse.urlparse(playlist_uri)

        if (
            export_uri_components.scheme != track_uri_components.scheme
            or export_uri_components.netloc != track_uri_components.netloc
        ):
            # save files to playlist on different location
            # return track path as is
            if track_uri_components.scheme == 'file':
                # as path if local file
                return Gio.File.new_for_uri(track_uri).get_path()
            return track_uri

        if options is None or not options.relative:
            # return absolute uri
            if track_uri_components.scheme == 'file':
                # as path if local file
                return Gio.File.new_for_uri(track_uri).get_path()
            return track_uri

        # calculate relative path
        track_uri = Gio.File.new_for_uri(track_uri)
        playlist_file_folder = Gio.File.new_for_uri(playlist_uri).get_parent()

        if track_uri_components.scheme == 'file':
            # return path
            # on Windows relpath raises a ValueError if track and playlist are not on same drive
            try:
                track_uri = os.path.relpath(
                    track_uri.get_path(), playlist_file_folder.get_path()
                )
                # Ensure '\' is always used as separator on Windows.
                if sys.platform == 'win32' and os.sep == '/':
                    track_uri = track_uri.replace(os.sep, '\\')
            except ValueError:
                return track_uri.get_path()
            return track_uri

        try:
            track_uri = os.path.relpath(
                track_uri.get_uri(), playlist_file_folder.get_uri()
            )
        except ValueError:
            return track_uri.get_uri()

        return track_uri


class M3UConverter(FormatConverter):
    """
    Import from and export to M3U format
    """

    title = _('M3U Playlist')
    content_types = ['audio/x-mpegurl', 'audio/mpegurl']

    def __init__(self):
        FormatConverter.__init__(self, 'm3u')

    def export_to_file(self, playlist, path, options=None):
        """
        Export a playlist to a given path

        :param playlist: the playlist
        :type playlist: :class:`Playlist`
        :param path: the target path
        :type path: string
        :param options: exporting options
        :type options: :class:`PlaylistExportOptions`
        """
        with GioFileOutputStream(Gio.File.new_for_uri(path)) as stream:
            stream.write('#EXTM3U\n')

            if playlist.name:
                stream.write('#PLAYLIST: {name}\n'.format(name=playlist.name))

            for track in playlist:
                title = [
                    track.get_tag_display('title', join=True),
                    track.get_tag_display('artist', join=True),
                ]
                length = int(round(float(track.get_tag_raw('__length') or -1)))

                track_path = track.get_loc_for_io()
                track_path = self.get_track_export_path(path, track_path, options)

                stream.write(
                    '#EXTINF:{length},{title}\n{path}\n'.format(
                        length=length,
                        title=' - '.join(title),
                        path=track_path,
                    )
                )

    def import_from_file(self, path):
        """
        Import a playlist from a given path

        :param path: the source path
        :type path: string
        :returns: the playlist
        :rtype: :class:`Playlist`
        """
        playlist = Playlist(name=self.name_from_path(path))
        extinf = {}
        lineno = 0

        logger.debug('Importing M3U playlist: %s', path)

        with GioFileInputStream(Gio.File.new_for_uri(path)) as stream:
            for line in stream:
                lineno += 1

                line = line.strip()

                if not line:
                    continue

                if line.upper().startswith('#PLAYLIST: '):
                    playlist.name = line[len('#PLAYLIST: ') :]
                elif line.startswith('#EXTINF:'):
                    extinf_line = line[len('#EXTINF:') :]

                    parts = extinf_line.split(',', 1)
                    length = 0

                    if len(parts) > 1 and int(parts[0]) > 0:
                        length = int(float(parts[0]))

                    extinf['__length'] = length

                    parts = parts[-1].rsplit(' - ', 1)

                    extinf['title'] = parts[-1]

                    if len(parts) > 1:
                        extinf['artist'] = parts[0]
                elif line.startswith('#'):
                    continue
                else:
                    track = trax.Track(self.get_track_import_path(path, line))

                    if extinf:
                        for tag, value in extinf.items():
                            if track.get_tag_raw(tag) is None:
                                try:
                                    track.set_tag_raw(tag, value)
                                except Exception as e:
                                    # Python 3: raise UnknownPlaylistTrackError() from e
                                    # Python 2: .. no good solution
                                    raise UnknownPlaylistTrackError(
                                        "line %s: %s" % (lineno, e)
                                    )

                    playlist.append(track)
                    extinf = {}

        return playlist


providers.register('playlist-format-converter', M3UConverter())


class PLSConverter(FormatConverter):
    """
    Import from and export to PLS format
    """

    title = _('PLS Playlist')
    content_types = ['audio/x-scpls']

    def __init__(self):
        FormatConverter.__init__(self, 'pls')

    def export_to_file(self, playlist, path, options=None):
        """
        Export a playlist to a given path

        :param playlist: the playlist
        :type playlist: :class:`Playlist`
        :param path: the target path
        :type path: string
        :param options: exporting options
        :type options: :class:`PlaylistExportOptions`
        """
        from configparser import RawConfigParser

        pls_playlist = RawConfigParser()
        pls_playlist.optionxform = str  # Make case sensitive
        pls_playlist.add_section('playlist')
        pls_playlist.set('playlist', 'NumberOfEntries', len(playlist))

        for index, track in enumerate(playlist):
            position = index + 1
            title = [
                track.get_tag_display('title', join=True),
                track.get_tag_display('artist', join=True),
            ]
            length = max(-1, int(round(float(track.get_tag_raw('__length') or -1))))

            track_path = track.get_loc_for_io()
            track_path = self.get_track_export_path(path, track_path, options)

            pls_playlist.set('playlist', 'File%d' % position, track_path)
            pls_playlist.set('playlist', 'Title%d' % position, ' - '.join(title))
            pls_playlist.set('playlist', 'Length%d' % position, length)

        pls_playlist.set('playlist', 'Version', 2)

        with GioFileOutputStream(Gio.File.new_for_uri(path)) as stream:
            pls_playlist.write(stream)

    def import_from_file(self, path):
        """
        Import a playlist from a given path

        :param path: the source path
        :type path: string
        :returns: the playlist
        :rtype: :class:`Playlist`
        """
        from configparser import (
            RawConfigParser,
            MissingSectionHeaderError,
            NoOptionError,
        )

        pls_playlist = RawConfigParser()
        gfile = Gio.File.new_for_uri(path)

        logger.debug('Importing PLS playlist: %s', path)

        try:
            with GioFileInputStream(gfile) as stream:
                pls_playlist.read_file(stream)
        except MissingSectionHeaderError:
            # Most likely version 1, thus only a list of URIs
            playlist = Playlist(self.name_from_path(path))

            with GioFileInputStream(gfile) as stream:
                for line in stream:
                    line = line.strip()

                    if not line:
                        continue

                    track = trax.Track(self.get_track_import_path(path, line))

                    if track.get_tag_raw('title') is None:
                        track.set_tag_raw(
                            'title', common.sanitize_url(self.name_from_path(line))
                        )

                    playlist.append(track)

            return playlist

        if not pls_playlist.has_section('playlist'):
            raise InvalidPlaylistTypeError(_('Invalid format for %s.') % self.title)

        if not pls_playlist.has_option('playlist', 'version'):
            logger.warning('No PLS version specified, ' 'assuming 2. [%s]', path)
            pls_playlist.set('playlist', 'version', 2)

        version = pls_playlist.getint('playlist', 'version')

        if version != 2:
            raise InvalidPlaylistTypeError(
                _('Unsupported version %(version)s for %(type)s')
                % {'version': version, 'type': self.title}
            )

        if not pls_playlist.has_option('playlist', 'numberofentries'):
            raise InvalidPlaylistTypeError(_('Invalid format for %s.') % self.title)

        # PLS playlists store no name, thus retrieve from path
        playlist = Playlist(common.sanitize_url(self.name_from_path(path)))
        numberofentries = pls_playlist.getint('playlist', 'numberofentries')

        for position in range(1, numberofentries + 1):
            try:
                uri = pls_playlist.get('playlist', 'file%d' % position)
            except NoOptionError:
                continue

            track = trax.Track(self.get_track_import_path(path, uri))
            title = artist = None
            length = 0

            try:
                title = pls_playlist.get('playlist', 'title%d' % position)
            except NoOptionError:
                title = common.sanitize_url(self.name_from_path(uri))
            else:
                title = title.split(' - ', 1)

                if len(title) > 1:  # "Artist - Title"
                    artist, title = title
                else:
                    title = title[0]

            try:
                length = pls_playlist.getint('playlist', 'length%d' % position)
            except NoOptionError:
                pass

            if track.get_tag_raw('title') is None and title:
                track.set_tag_raw('title', title)

            if track.get_tag_raw('artist') is None and artist:
                track.set_tag_raw('artist', artist)

            if track.get_tag_raw('__length') is None:
                track.set_tag_raw('__length', max(0, length))

            playlist.append(track)

        return playlist


providers.register('playlist-format-converter', PLSConverter())


class ASXConverter(FormatConverter):
    """
    Import from and export to ASX format
    """

    title = _('ASX Playlist')
    content_types = [
        'video/x-ms-asf',
        'audio/x-ms-asx',
        'audio/x-ms-wax',
        'video/x-ms-wvx',
    ]
    file_extensions = ['asx', 'wax', 'wvx']

    def __init__(self):
        FormatConverter.__init__(self, 'asx')

    def export_to_file(self, playlist, path, options=None):
        """
        Export a playlist to a given path

        :param playlist: the playlist
        :type playlist: :class:`Playlist`
        :param path: the target path
        :type path: string
        :param options: exporting options
        :type options: :class:`PlaylistExportOptions`
        """
        from xml.sax.saxutils import escape

        with GioFileOutputStream(Gio.File.new_for_uri(path)) as stream:
            stream.write('<asx version="3.0">\n')
            stream.write('  <title>%s</title>\n' % escape(playlist.name))

            for track in playlist:
                stream.write('  <entry>\n')

                title = track.get_tag_raw('title', join=True)
                artist = track.get_tag_raw('artist', join=True)

                if title:
                    stream.write('    <title>%s</title>\n' % escape(title))

                if artist:
                    stream.write('    <author>%s</author>\n' % escape(artist))

                track_path = track.get_loc_for_io()
                track_path = self.get_track_export_path(path, track_path, options)

                stream.write('    <ref href="%s" />\n' % track_path)
                stream.write('  </entry>\n')

            stream.write('</asx>')

    def import_from_file(self, path):
        """
        Import a playlist from a given path

        :param path: the source path
        :type path: string
        :returns: the playlist
        :rtype: :class:`Playlist`
        """
        from xml.etree.cElementTree import XMLParser

        playlist = Playlist(self.name_from_path(path))

        logger.debug('Importing ASX playlist: %s', path)

        with GioFileInputStream(Gio.File.new_for_uri(path)) as stream:
            parser = XMLParser(target=self.ASXPlaylistParser())
            parser.feed(stream.read())

            try:
                playlistdata = parser.close()
            except Exception:
                pass
            else:
                if playlistdata['name']:
                    playlist.name = playlistdata['name']

                for trackdata in playlistdata['tracks']:
                    track = trax.Track(
                        self.get_track_import_path(path, trackdata['uri'])
                    )

                    ntags = {}
                    for tag, value in trackdata['tags'].items():
                        if not track.get_tag_raw(tag) and value:
                            ntags[tag] = value
                    if ntags:
                        track.set_tags(**ntags)

                    playlist.append(track)

        return playlist

    class ASXPlaylistParser:
        """
        Target for xml.etree.ElementTree.XMLParser, allows
        for parsing ASX playlists case-insensitive
        """

        def __init__(self):
            self._stack = deque()

            self._playlistdata = {'name': None, 'tracks': []}
            self._trackuri = None
            self._trackdata = {}

        def start(self, tag, attributes):
            """
            Checks the ASX version and stores
            the URI of the current track
            """
            depth = len(self._stack)
            # Convert both tag and attributes to lowercase
            tag = tag.lower()
            attributes = {k.lower(): v for k, v in attributes.items()}

            if depth > 0:
                if depth == 2 and self._stack[-1] == 'entry' and tag == 'ref':
                    self._trackuri = attributes.get('href', None)
            # Check root element and version
            elif tag != 'asx' or attributes.get('version', None) != '3.0':
                return

            self._stack.append(tag)

        def data(self, data):
            """
            Stores track data and playlist name
            """
            depth = len(self._stack)

            if depth > 0 and data:
                element = self._stack[-1]

                if depth == 3:
                    # Only consider title and author for now
                    if element == 'title':
                        self._trackdata['title'] = data
                    elif element == 'author':
                        self._trackdata['artist'] = data
                elif depth == 2 and element == 'title':
                    self._playlistdata['name'] = data

        def end(self, tag):
            """
            Appends track data
            """
            try:
                self._stack.pop()
            except IndexError:  # Invalid playlist
                pass
            else:
                if tag.lower() == 'entry':
                    # Only add track data if we have at least an URI
                    if self._trackuri:
                        self._playlistdata['tracks'].append(
                            {'uri': self._trackuri, 'tags': self._trackdata.copy()}
                        )

                    self._trackuri = None
                    self._trackdata.clear()

        def close(self):
            """
            Returns the playlist data including
            data of all successfully read tracks

            :rtype: dict
            """
            return self._playlistdata


providers.register('playlist-format-converter', ASXConverter())


class XSPFConverter(FormatConverter):
    """
    Import from and export to XSPF format
    """

    title = _('XSPF Playlist')
    content_types = ['application/xspf+xml']

    def __init__(self):
        FormatConverter.__init__(self, 'xspf')

        # TODO: support image tag for CoverManager
        self.tags = {
            'title': 'title',
            'creator': 'artist',
            'album': 'album',
            'trackNum': 'tracknumber',
        }

    def export_to_file(self, playlist, path, options=None):
        """
        Export a playlist to a given path

        :param playlist: the playlist
        :type playlist: :class:`Playlist`
        :param path: the target path
        :type path: string
        :param options: exporting options
        :type options: :class:`PlaylistExportOptions`
        """
        from xml.sax.saxutils import escape

        with GioFileOutputStream(Gio.File.new_for_uri(path)) as stream:
            stream.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            stream.write('<playlist version="1" xmlns="http://xspf.org/ns/0/">\n')

            if playlist.name:
                stream.write('  <title>%s</title>\n' % escape(playlist.name))

            stream.write('  <trackList>\n')

            for track in playlist:
                stream.write('    <track>\n')
                for element, tag in self.tags.items():
                    if not track.get_tag_raw(tag):
                        continue
                    stream.write(
                        '      <%s>%s</%s>\n'
                        % (element, escape(track.get_tag_raw(tag, join=True)), element)
                    )

                track_path = track.get_loc_for_io()
                track_path = self.get_track_export_path(path, track_path, options)

                stream.write('      <location>%s</location>\n' % escape(track_path))
                stream.write('    </track>\n')

            stream.write('  </trackList>\n')
            stream.write('</playlist>\n')

    def import_from_file(self, path):
        """
        Import a playlist from a given path

        :param path: the source path
        :type path: string
        :returns: the playlist
        :rtype: :class:`Playlist`
        """
        # TODO: support content resolution
        import xml.etree.cElementTree as ETree

        playlist = Playlist(name=self.name_from_path(path))

        logger.debug('Importing XSPF playlist: %s', path)

        with GioFileInputStream(Gio.File.new_for_uri(path)) as stream:
            tree = ETree.ElementTree(file=stream)
            ns = "{http://xspf.org/ns/0/}"
            nodes = tree.find("%strackList" % ns).findall("%strack" % ns)
            titlenode = tree.find("%stitle" % ns)

            if titlenode is not None:
                playlist.name = titlenode.text.strip()

            for n in nodes:
                location = n.find("%slocation" % ns).text.strip()
                track = trax.Track(self.get_track_import_path(path, location))
                for element, tag in self.tags.items():
                    try:
                        track.set_tag_raw(
                            tag, n.find("%s%s" % (ns, element)).text.strip()
                        )
                    except Exception:
                        pass
                playlist.append(track)

        return playlist


providers.register('playlist-format-converter', XSPFConverter())


class Playlist:
    # TODO: how do we document events in sphinx?
    """
    Basic class for handling a list of tracks

    EVENTS: (all events are synchronous)
        * playlist_tracks_added
            * fired: after tracks are added
            * data: list of tuples of (index, track)
        * playlist_tracks_removed
            * fired: after tracks are removed
            * data: list of tuples of (index, track)
        * playlist_current_position_changed
        * playlist_shuffle_mode_changed
        * playlist_random_mode_changed
        * playlist_dynamic_mode_changed
    """
    #: Valid shuffle modes (list of string)
    shuffle_modes = ['disabled', 'track', 'album', 'random']
    #: Titles of the valid shuffle modes (list of string)
    shuffle_mode_names = [
        _('Shuffle _Off'),
        _('Shuffle _Tracks'),
        _('Shuffle _Albums'),
        _('_Random'),
    ]
    #: Valid repeat modes (list of string)
    repeat_modes = ['disabled', 'all', 'track']
    #: Titles of the valid repeat modes (list of string)
    repeat_mode_names = [_('Repeat _Off'), _('Repeat _All'), _('Repeat O_ne')]
    #: Valid dynamic modes
    dynamic_modes = ['disabled', 'enabled']
    #: Titles of the valid dynamic modes
    dynamic_mode_names = [_('Dynamic _Off'), _('Dynamic by Similar _Artists')]
    save_attrs = [
        'shuffle_mode',
        'repeat_mode',
        'dynamic_mode',
        'current_position',
        'name',
    ]
    __playlist_format_version = [2, 0]

    def __init__(self, name, initial_tracks=[]):
        """
        :param name: the initial name of the playlist
        :type name: string
        :param initial_tracks: the tracks which shall
            populate the playlist initially
        :type initial_tracks: list of :class:`xl.trax.Track`
        """
        self.__tracks = MetadataList()
        for track in initial_tracks:
            if not isinstance(track, trax.Track):
                raise ValueError("Need trax.Track object, got %r" % type(track))
            self.__tracks.append(track)
        self.__shuffle_mode = self.shuffle_modes[0]
        self.__repeat_mode = self.repeat_modes[0]
        self.__dynamic_mode = self.dynamic_modes[0]

        # dirty: any change that would alter the on-disk
        #   Representation should set this.
        # needs_save: changes to list content should set this.
        #   Determines when the 'unsaved' indicator is shown to the user.
        # next_data: cached info to determine next track.
        #   None or a tuple of (spat, index, next):
        #     spat: <bool> (is SPAT activated on the current track?)
        #     index: <int> index in self.__tracks or None
        #     next: <xl.trax.Track> or None
        # current_position: <int> index in self.__tracks or -1 if no track
        # spat_position: <int> index in self.__tracks or -1 if no SPAT set
        # shuffle_history_counter: <int> count of tracks queued in shuffle mode
        #   Start positive so we can just do an if directly on the value.
        self.__dirty = False
        self.__needs_save = False
        self.__name = name
        self.__next_data = None
        self.__current_position = -1
        self.__spat_position = -1
        self.__shuffle_history_counter = 1

        event.add_callback(self.on_playback_track_start, "playback_track_start")

    ### playlist-specific API ###

    def _set_name(self, name):
        self.__name = name
        self.__needs_save = self.__dirty = True
        event.log_event("playlist_name_changed", self, name)

    #: The playlist name (string)
    name = property(lambda self: self.__name, _set_name)
    #: Whether the playlist was changed or not (boolean)
    dirty = property(lambda self: self.__dirty)

    def clear(self):
        """
        Removes all contained tracks
        """
        del self[:]

    def get_current_position(self):
        """
        Retrieves the current position within the playlist

        :returns: the position
        :rtype: int
        """
        return self.__current_position

    def set_current_position(self, position):
        """
        Sets the current position within the playlist

        :param position: the new position
        :type position: int
        """
        self.__next_data = None
        oldposition = self.__current_position
        if oldposition == position:
            return
        if position != -1:
            if position >= len(self.__tracks):
                raise IndexError("Cannot set position past end of playlist")
            self.__tracks.set_meta_key(position, "playlist_current_position", True)
        self.__current_position = position
        if oldposition != -1:
            try:
                self.__tracks.del_meta_key(oldposition, "playlist_current_position")
            except KeyError:
                pass
        self.__dirty = True
        event.log_event(
            "playlist_current_position_changed", self, (position, oldposition)
        )

    #: The position within the playlist (int)
    current_position = property(get_current_position, set_current_position)

    def get_spat_position(self):
        """
        Retrieves the current position within the playlist
        after which progressing shall be stopped

        :returns: the position
        :rtype: int
        """
        return self.__spat_position

    def set_spat_position(self, position):
        """
        Sets the current position within the playlist
        after which progressing shall be stopped

        :param position: the new position
        :type position: int
        """
        self.__next_data = None
        oldposition = self.spat_position
        self.__tracks.set_meta_key(position, "playlist_spat_position", True)
        self.__spat_position = position
        if oldposition != -1:
            try:
                self.__tracks.del_meta_key(oldposition, "playlist_spat_position")
            except KeyError:
                pass
        self.__dirty = True
        event.log_event("playlist_spat_position_changed", self, (position, oldposition))

    #: The position within the playlist after which to stop progressing (int)
    spat_position = property(get_spat_position, set_spat_position)

    def get_current(self):
        """
        Retrieves the track at the current position

        :returns: the track
        :rtype: :class:`xl.trax.Track` or None
        """
        if self.current_position == -1:
            return None
        return self.__tracks[self.current_position]

    current = property(get_current)

    def get_shuffle_history(self):
        """
        Retrieves the history of played
        tracks from a shuffle run

        :returns: the tracks
        :rtype: list
        """
        return [
            (i, self.__tracks[i])
            for i in range(len(self))
            if self.__tracks.get_meta_key(i, 'playlist_shuffle_history')
        ]

    def clear_shuffle_history(self):
        """
        Clear the history of played
        tracks from a shuffle run
        """
        for i in range(len(self)):
            try:
                self.__tracks.del_meta_key(i, "playlist_shuffle_history")
            except Exception:
                pass

    @common.threaded
    def __fetch_dynamic_tracks(self):
        dynamic.MANAGER.populate_playlist(self)

    def __next_random_track(self, current_position, mode="track"):
        """
        Returns a valid next track if shuffle is activated based
        on random_mode
        """
        if mode == "album":
            # TODO: we really need proper album-level operations in
            # xl.trax for this
            try:
                # Try and get the next track on the album
                # NB If the user starts the playlist from the middle
                # of the album some tracks of the album remain off the
                # tracks_history, and the album can be selected again
                # randomly from its first track
                if current_position == -1:
                    raise IndexError
                curr = self[current_position]
                t = [
                    x
                    for i, x in enumerate(self)
                    if x.get_tag_raw('album') == curr.get_tag_raw('album')
                    and i > current_position
                ]
                t = trax.sort_tracks(['discnumber', 'tracknumber'], t)
                return self.__tracks.index(t[0]), t[0]

            except IndexError:  # Pick a new album
                hist = set(self.get_shuffle_history())
                albums = set()
                for i, x in enumerate(self):
                    if (i, x) in hist:
                        continue
                    alb = x.get_tag_raw('album')
                    if alb:
                        albums.add(tuple(alb))
                if not albums:
                    return -1, None
                album = list(random.choice(list(albums)))
                t = [x for x in self if x.get_tag_raw('album') == album]
                t = trax.sort_tracks(['tracknumber'], t)
                return self.__tracks.index(t[0]), t[0]
        elif mode == 'random':
            try:
                return random.choice(
                    [(i, self.__tracks[i]) for i, tr in enumerate(self.__tracks)]
                )
            except IndexError:
                return -1, None
        else:
            hist = {i for i, tr in self.get_shuffle_history()}
            try:
                return random.choice(
                    [
                        (i, self.__tracks[i])
                        for i, tr in enumerate(self.__tracks)
                        if i not in hist
                    ]
                )
            except IndexError:  # no more tracks
                return -1, None

    def __get_next(self, current_position):
        # don't recalculate
        if self.__next_data is not None:
            return self.__next_data[2]

        repeat_mode = self.repeat_mode
        shuffle_mode = self.shuffle_mode
        next_index = -1
        spat = (
            True
            if current_position == self.spat_position and current_position != -1
            else False
        )

        if repeat_mode == 'track':
            next_index = None
            next = self.current
        elif shuffle_mode != 'disabled':
            if current_position != -1:
                self.__tracks.set_meta_key(
                    current_position,
                    "playlist_shuffle_history",
                    self.__shuffle_history_counter,
                )
                self.__shuffle_history_counter += 1
            next_index, next = self.__next_random_track(current_position, shuffle_mode)
            if next is None:
                self.clear_shuffle_history()
        else:
            try:
                next = self[current_position + 1]
                next_index = current_position + 1
            except IndexError:
                next = None

        if next is None:
            if repeat_mode == 'all':
                if len(self) == 1:
                    next = self[current_position]
                    next_index = current_position
                if len(self) > 1:
                    return self.__get_next(-1)

        self.__next_data = (spat, next_index, next)
        return next

    def get_next(self):
        """
        Retrieves the next track that will be played. Does not
        actually set the position. When you call next(), it should
        return the same track, even in random shuffle modes.

        This exists to support retrieving a track before it actually
        needs to be played, such as for pre-buffering.

        :returns: the next track to be played
        :rtype: :class:`xl.trax.Track` or None
        """
        return self.__get_next(self.current_position)

    def next(self):
        """
        Progresses to the next track within the playlist
        and takes shuffle and repeat modes into account

        :returns: the new current track
        :rtype: :class:`xl.trax.Track` or None
        """

        if not self.__next_data:
            self.__get_next(self.current_position)

        spat, index, next = self.__next_data

        if index is not None:
            try:
                self.current_position = index
            except IndexError:
                self.current_position = -1
        else:
            self.__next_data = None
        if spat:
            self.spat_position = -1
            return None

        return self.current

    def prev(self):
        """
        Progresses to the previous track within the playlist
        and takes shuffle and repeat modes into account

        :returns: the new current track
        :rtype: :class:`xl.trax.Track` or None
        """
        repeat_mode = self.repeat_mode
        shuffle_mode = self.shuffle_mode
        if repeat_mode == 'track':
            return self.current

        if shuffle_mode != 'disabled':
            shuffle_hist, prev_index = max(
                (self.__tracks.get_meta_key(i, 'playlist_shuffle_history', 0), i)
                for i in range(len(self))
            )

            if shuffle_hist:
                self.current_position = prev_index
                self.__tracks.del_meta_key(prev_index, 'playlist_shuffle_history')
        else:
            position = self.current_position - 1
            if position < 0:
                if repeat_mode == 'all':
                    position = len(self) - 1
                else:
                    position = 0 if len(self) else -1
            self.current_position = position
        return self.get_current()

    ### track advance modes ###
    # This code may look a little overkill, but it's this way to
    # maximize forwards-compatibility. get_ methods will not overwrite
    # currently-set modes which may be from a future version, while set_
    # methods explicitly disallow modes not supported in this version.
    # This ensures that 1) saved modes are never clobbered unless a
    # known mode is to be set, and 2) the values returned in _mode will
    # always be supported in the running version.

    def __get_mode(self, modename):
        mode = getattr(self, "_Playlist__%s_mode" % modename)
        modes = getattr(self, "%s_modes" % modename)
        if mode in modes:
            return mode
        else:
            return modes[0]

    def __set_mode(self, modename, mode):
        modes = getattr(self, "%s_modes" % modename)
        if mode not in modes:
            raise TypeError("Mode %s is invalid" % mode)
        else:
            self.__dirty = True
            setattr(self, "_Playlist__%s_mode" % modename, mode)
            event.log_event("playlist_%s_mode_changed" % modename, self, mode)

    def get_shuffle_mode(self):
        """
        Retrieves the current shuffle mode

        :returns: the shuffle mode
        :rtype: string
        """
        return self.__get_mode("shuffle")

    def set_shuffle_mode(self, mode):
        """
        Sets the current shuffle mode

        :param mode: the new shuffle mode
        :type mode: string
        """
        self.__set_mode("shuffle", mode)
        if mode == 'disabled':
            self.clear_shuffle_history()

    #: The current shuffle mode (string)
    shuffle_mode = property(get_shuffle_mode, set_shuffle_mode)

    def get_repeat_mode(self):
        """
        Retrieves the current repeat mode

        :returns: the repeat mode
        :rtype: string
        """
        return self.__get_mode('repeat')

    def set_repeat_mode(self, mode):
        """
        Sets the current repeat mode

        :param mode: the new repeat mode
        :type mode: string
        """
        self.__set_mode("repeat", mode)

    #: The current repeat mode (string)
    repeat_mode = property(get_repeat_mode, set_repeat_mode)

    def get_dynamic_mode(self):
        """
        Retrieves the current dynamic mode

        :returns: the dynamic mode
        :rtype: string
        """
        return self.__get_mode("dynamic")

    def set_dynamic_mode(self, mode):
        """
        Sets the current dynamic mode

        :param mode: the new dynamic mode
        :type mode: string
        """
        self.__set_mode("dynamic", mode)

    #: The current dynamic mode (string)
    dynamic_mode = property(get_dynamic_mode, set_dynamic_mode)

    def randomize(self, positions=None):
        """
        Randomizes the content of the playlist contrary to
        shuffle which affects only the progressing order

        By default all tracks in the playlist are randomized,
        but a list of positions can be passed. The tracks on
        these positions will be randomized, all other tracks
        will keep their positions.

        :param positions: list of track positions to randomize
        :type positions: iterable
        """
        # Turn 2 lists into a list of tuples
        tracks = list(zip(self.__tracks, self.__tracks.metadata))

        if positions:
            # For 2 items, simple swapping is most reasonable
            if len(positions) == 2:
                tracks[positions[0]], tracks[positions[1]] = (
                    tracks[positions[1]],
                    tracks[positions[0]],
                )
            else:
                # Extract items and shuffle them
                shuffle_tracks = [t for i, t in enumerate(tracks) if i in positions]
                random.shuffle(shuffle_tracks)

                # Put shuffled items back
                for position in positions:
                    tracks[position] = shuffle_tracks.pop()
        else:
            random.shuffle(tracks)

        # Turn list of tuples into 2 tuples
        self[:] = MetadataList(*zip(*tracks))

    def sort(self, tags, reverse=False):
        """
        Sorts the content of the playlist

        :param tags: tags to sort by
        :type tags: list of strings
        :param reverse: whether the sorting shall be reversed
        :type reverse: boolean
        """
        data = zip(self.__tracks, self.__tracks.metadata)
        data = trax.sort_tracks(
            tags, data, trackfunc=operator.itemgetter(0), reverse=reverse
        )
        l = MetadataList()
        l.extend([x[0] for x in data])
        l.metadata = [x[1] for x in data]
        self[:] = l

    # TODO[0.4?]: drop our custom disk playlist format in favor of an
    # extended XSPF playlist (using xml namespaces?).

    # TODO: add timeout saving support. 5-10 seconds after last change,
    # perhaps?

    def save_to_location(self, location):
        """
        Writes the content of the playlist to a given location

        :param location: the location to save to
        :type location: string
        """
        new_location = location + ".new"

        with open(new_location, 'w') as f:
            for track in self.__tracks:
                loc = track.get_loc_for_io()
                meta = {}
                for tag in ('artist', 'album', 'tracknumber', 'title', 'genre', 'date'):
                    value = track.get_tag_raw(tag, join=True)
                    if value is not None:
                        meta[tag] = value
                meta = urllib.parse.urlencode(meta)
                print(loc, meta, sep='\t', file=f)

            print('EOF', file=f)

            for attr in self.save_attrs:
                val = getattr(self, attr)
                try:
                    configstr = settings.MANAGER._val_to_str(val)
                except ValueError:
                    configstr = ''
                print('%s=%s' % (attr, configstr), file=f)

        os.replace(new_location, location)

        self.__needs_save = self.__dirty = False

    def load_from_location(self, location):
        """
        Loads the content of the playlist from a given location

        :param location: the location to load from
        :type location: string
        """
        # note - this is not guaranteed to fire events when it sets
        # attributes. It is intended ONLY for initial setup, not for
        # reloading a playlist inline.
        f = None
        for loc in [location, location + ".new"]:
            try:
                f = open(loc, 'r')
                break
            except Exception:
                pass
        if not f:
            return
        locs = []
        while True:
            line = f.readline()
            if line == "EOF\n" or line == "":
                break
            locs.append(line.strip())
        items = {}
        while True:
            line = f.readline()
            if line == "":
                break

            try:
                item, strn = line[:-1].split("=", 1)
            except ValueError:
                continue  # Skip erroneous lines

            val = settings.MANAGER._str_to_val(strn)
            items[item] = val

        ver = items.get("__playlist_format_version", [1])
        if ver[0] == 1:
            if items.get("repeat_mode") == "playlist":
                items['repeat_mode'] = "all"
        elif ver[0] > self.__playlist_format_version[0]:
            raise IOError("Cannot load playlist, unknown format")
        elif ver > self.__playlist_format_version:
            logger.warning(
                "Playlist created on a newer Exaile version, some attributes may not be handled."
            )
        f.close()

        trs = []

        for loc in locs:
            meta = None
            if loc.find('\t') > -1:
                splitted = loc.split('\t')
                loc = "\t".join(splitted[:-1])
                meta = splitted[-1]

            track = None
            track = trax.Track(uri=loc)

            # re-add meta
            if not track:
                continue
            if not track.is_local() and meta is not None:
                meta = urllib.parse.parse_qs(meta)
                for k, v in meta.items():
                    track.set_tag_raw(k, v[0], notify_changed=False)

            trs.append(track)

        self.__tracks[:] = trs

        for item, val in items.items():
            if item in self.save_attrs:
                try:
                    setattr(self, item, val)
                except TypeError:  # don't bail if we try to set an invalid mode
                    logger.debug(
                        "Got a TypeError when trying to set attribute %s to %s during playlist restore.",
                        item,
                        val,
                    )

    def reverse(self):
        # reverses current view
        pass

    ### list-like API methods ###
    # parts of this section are taken from
    # https://code.activestate.com/recipes/440656-list-mixin/

    def __len__(self):
        return len(self.__tracks)

    def __contains__(self, track):
        return track in self.__tracks

    def __tuple_from_slice(self, i):
        """
        Get (start, end, step) tuple from slice object.
        """
        start, end, step = i.indices(len(self))
        if i.step is None:
            step = 1
        return (start, end, step)

    def __adjust_current_pos(self, oldpos, removed, added):
        newpos = oldpos
        for i, tr in removed:
            if i <= oldpos:
                newpos -= 1
        for i, tr in added:
            if i <= newpos:
                newpos += 1
        self.current_position = newpos

    def __getitem__(self, i):
        return self.__tracks.__getitem__(i)

    def __setitem__(self, i, value):
        oldtracks = self.__getitem__(i)
        removed = MetadataList()
        added = MetadataList()
        oldpos = self.current_position

        if isinstance(i, slice):
            for x in value:
                if not isinstance(x, trax.Track):
                    raise ValueError("Need trax.Track object, got %r" % type(x))

            start, end, step = self.__tuple_from_slice(i)

            if isinstance(value, MetadataList):
                metadata = value.metadata
            else:
                metadata = [None] * len(value)

            if step != 1:
                if len(value) != len(oldtracks):
                    raise ValueError("Extended slice assignment must match sizes.")
            self.__tracks.__setitem__(i, value)
            removed = MetadataList(
                zip(range(start, end, step), oldtracks), oldtracks.metadata
            )
            if step == 1:
                end = start + len(value)

            added = MetadataList(zip(range(start, end, step), value), metadata)
        else:
            if not isinstance(value, trax.Track):
                raise ValueError("Need trax.Track object, got %r" % type(value))
            self.__tracks[i] = value
            removed = [(i, oldtracks)]
            added = [(i, value)]

        self.on_tracks_changed()

        if removed:
            event.log_event('playlist_tracks_removed', self, removed)
        if added:
            event.log_event('playlist_tracks_added', self, added)
        self.__adjust_current_pos(oldpos, removed, added)

        self.__needs_save = self.__dirty = True

    def __delitem__(self, i):
        if isinstance(i, slice):
            start, end, step = self.__tuple_from_slice(i)
        oldtracks = self.__getitem__(i)
        oldpos = self.current_position
        self.__tracks.__delitem__(i)
        removed = MetadataList()

        if isinstance(i, slice):
            removed = MetadataList(
                zip(range(start, end, step), oldtracks), oldtracks.metadata
            )
        else:
            removed = [(i, oldtracks)]

        self.on_tracks_changed()
        event.log_event('playlist_tracks_removed', self, removed)
        self.__adjust_current_pos(oldpos, removed, [])
        self.__needs_save = self.__dirty = True

    def append(self, other):
        """
        Appends a single track to the playlist

        Prefer extend() for batch updates, so that
        playlist_tracks_added is not emitted excessively.

        :param other: a :class:`xl.trax.Track`
        """
        self[len(self) : len(self)] = [other]

    def extend(self, other):
        """
        Extends the playlist by another playlist

        :param other: list of :class:`xl.trax.Track`
        """
        self[len(self) : len(self)] = other

    def count(self, other):
        """
        Returns the count of contained tracks

        :returns: the count
        :rtype: int
        """
        return self.__tracks.count(other)

    def index(self, item, start=0, end=None):
        """
        Retrieves the index of a track within the playlist

        :returns: the index
        :rtype: int
        """
        if end is None:
            return self.__tracks.index(item, start)
        else:
            return self.__tracks.index(item, start, end)

    def pop(self, i=-1):
        """
        Pops a track from the playlist

        :param i: the index
        :type i: int
        :returns: the track
        :rtype: :class:`xl.trax.Track`
        """
        item = self[i]
        del self[i]
        return item

    def on_playback_track_start(self, event_type, player, track):
        if player.queue is not None and player.queue.current_playlist == self:
            if self.dynamic_mode != 'disabled':
                self.__fetch_dynamic_tracks()

    def on_tracks_changed(self, *args):
        for idx in range(len(self.__tracks)):
            if self.__tracks.get_meta_key(idx, "playlist_current_position"):
                self.__current_position = idx
                break
        else:
            self.__current_position = -1
        for idx in range(len(self.__tracks)):
            if self.__tracks.get_meta_key(idx, "playlist_spat_position"):
                self.__spat_position = idx
                break
        else:
            self.__spat_position = -1


class SmartPlaylist:
    """
    Represents a Smart Playlist.
    This will query a collection object using a set of parameters

    Simple usage:

    >>> import xl.collection
    >>> col = xl.collection.Collection("Test Collection")
    >>> col.add_library(xl.collection.Library("./tests/data"))
    >>> col.rescan_libraries()
    >>> sp = SmartPlaylist(collection=col)
    >>> sp.add_param("artist", "==", "Delerium")
    >>> p = sp.get_playlist()
    >>> p[1]['album'][0]
    'Chimera'
    >>>
    """

    def __init__(self, name="", collection=None):
        """
        Sets up a smart playlist

        @param collection: a reference to a TrackDB object.
        """
        self.search_params = []
        self.custom_params = []
        self.collection = collection
        self.or_match = False
        self.track_count = -1
        self.random_sort = False
        self.name = name
        self.sort_tags = None
        self.sort_order = None

    def set_location(self, location):
        pass

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    def set_collection(self, collection):
        """
        change the collection backing this playlist

        collection: the collection to use [Collection]
        """
        self.collection = collection

    def set_random_sort(self, sort):
        """
        If True, the tracks added during update() will be randomized

        @param sort: bool
        """
        self.random_sort = sort
        self._dirty = True

    def get_random_sort(self):
        """
        Returns True if this playlist will randomly be sorted
        """
        return self.random_sort

    def set_return_limit(self, count):
        """
        Sets the max number of tracks to return.

        @param count:  number of tracks to return.  Set to -1 to return
            all matched
        """
        self.track_count = count
        self._dirty = True

    def get_return_limit(self):
        """
        Returns the track count setting
        """
        return self.track_count

    def set_sort_tags(self, tags, reverse):
        """
        Control playlist sorting

        :param tags: List of tags to sort by
        :param reverse: Reverse the tracks after sorting
        """
        self.sort_tags = tags
        self.sort_order = reverse

    def get_sort_tags(self):
        """
        :returns: (list of tags, reverse)
        """
        return self.sort_tags, self.sort_order

    def set_or_match(self, value):
        """
        Set to True to make this an or match: match any of the
        parameters

        value: True to match any, False to match all params
        """
        self.or_match = value
        self._dirty = True

    def get_or_match(self):
        """
        Return if this is an any or and playlist
        """
        return self.or_match

    def add_param(self, field, op, value, index=-1):
        """
        Adds a search parameter.

        @param field:  The field to operate on. [string]
        @param op:     The operator.  Valid operators are:
                >,<,>=,<=,=,!=,==,!==,>< (between) [string]
        @param value:  The value to match against [string]
        @param index:  Where to insert the parameter in the search
                order.  -1 to append [int]
        """
        if index:
            self.search_params.insert(index, [field, op, value])
        else:
            self.search_params.append([field, op, value])
        self._dirty = True

    def set_custom_param(self, param, index=-1):
        """
        Adds an arbitrary search parameter, exposing the full power
        of the new search system to the user.

        param:  the search query to use. [string]
        index:  the index to insert at. default is append [int]
        """
        if index:
            self.search_params.insert(index, param)
        else:
            self.search_params.append(param)
        self._dirty = True

    def remove_param(self, index):
        """
        Removes a parameter at the specified index

        index:  the index of the parameter to remove
        """
        self._dirty = True
        return self.search_params.pop(index)

    def get_playlist(self, collection=None):
        """
        Generates a playlist by querying the collection

        @param collection: the collection to search (leave None to
                search internal ref)
        """
        pl = Playlist(name=self.name)
        if not collection:
            collection = self.collection
        if not collection:  # if there wasn't one set we might not have one
            return pl

        search_string, matchers = self._create_search_data(collection)

        matcher = trax.TracksMatcher(search_string, case_sensitive=False)

        # prepend for now, since it is likely to remove more tracks, and
        # smart playlists don't support mixed and/or expressions yet
        for m in matchers:
            matcher.prepend_matcher(m, self.or_match)

        trs = [t.track for t in trax.search_tracks(collection, [matcher])]
        if self.random_sort:
            random.shuffle(trs)
        else:
            order = False
            if self.sort_tags:
                order = self.sort_order
                sort_by = [self.sort_tags] + list(common.BASE_SORT_TAGS)
            else:
                sort_by = common.BASE_SORT_TAGS
            trs = trax.sort_tracks(sort_by, trs, reverse=order)
        if self.track_count > 0 and len(trs) > self.track_count:
            trs = trs[: self.track_count]

        pl.extend(trs)

        return pl

    def _create_search_data(self, collection):
        """
        Creates a search string + matchers based on the internal params
        """

        params = []  # parameter list
        matchers = []  # matchers list
        maximum = settings.get_option('rating/maximum', 5)
        durations = {
            'seconds': lambda value: timedelta(seconds=value),
            'minutes': lambda value: timedelta(minutes=value),
            'hours': lambda value: timedelta(hours=value),
            'days': lambda value: timedelta(days=value),
            'weeks': lambda value: timedelta(weeks=value),
        }

        for param in self.search_params:
            if isinstance(param, str):
                params += [param]
                continue
            field, op, value = param
            fieldtype = tag_data.get(field)
            if fieldtype is not None:
                fieldtype = fieldtype.type

            s = ""

            if field == '__rating':
                value = 100.0 * value / maximum
            elif field == '__playlist':
                try:
                    pl = main.exaile().playlists.get_playlist(value)
                except Exception:
                    try:
                        pl = (
                            main.exaile()
                            .smart_playlists.get_playlist(value)
                            .get_playlist(collection)
                        )
                    except Exception as e:
                        raise ValueError("Loading %s: %s" % (self.name, e))

                if op == 'pin':
                    matchers.append(trax.TracksInList(pl))
                else:
                    matchers.append(trax.TracksNotInList(pl))
                continue
            elif fieldtype == 'timestamp':
                duration, unit = value
                delta = durations[unit](duration)
                point = datetime.now() - delta
                value = time.mktime(point.timetuple())

            # fmt: off
            if op == ">=" or op == "<=":
                s += '( %(field)s%(op)s%(value)s ' \
                    '| %(field)s==%(value)s )' % \
                    {
                        'field': field,
                        'value': value,
                        'op': op[0]
                    }
            elif op == "!=" or op == "!==" or op == "!~":
                s += '! %(field)s%(op)s"%(value)s"' % \
                    {
                        'field': field,
                        'value': value,
                        'op': op[1:]
                    }
            elif op == "><":
                s += '( %(field)s>%(value1)s ' \
                    '%(field)s<%(value2)s )' % \
                    {
                        'field': field,
                        'value1': value[0],
                        'value2': value[1]
                    }
            elif op == '<!==>':     # NOT NULL
                s += '! %(field)s=="__null__"' % \
                    {
                        'field': field
                    }
            elif op == '<==>':      # IS NULL
                s += '%(field)s=="__null__"' % \
                    {
                        'field': field
                    }
            elif op == 'w=': # contains word
                s += '%(field)s~"\\b%(value)s\\b"' % \
                    {
                        'field': field,
                        'value': re.escape(value),
                    }
            elif op == '!w=': # does not contain word
                s += '! %(field)s~"\\b%(value)s\\b"' % \
                    {
                        'field': field,
                        'value': re.escape(value),
                    }
            else:
                s += '%(field)s%(op)s"%(value)s"' % \
                    {
                        'field': field,
                        'value': value,
                        'op': op
                    }

            # fmt: on

            params.append(s)

        if self.or_match:
            return ' | '.join(params), matchers
        else:
            return ' '.join(params), matchers

    def save_to_location(self, location):
        pdata = {}
        for item in [
            'search_params',
            'custom_params',
            'or_match',
            'track_count',
            'random_sort',
            'name',
            'sort_tags',
            'sort_order',
        ]:
            pdata[item] = getattr(self, item)
        with open(location, 'wb') as fp:
            pickle.dump(pdata, fp)

    def load_from_location(self, location):
        try:
            with open(location, 'rb') as fp:
                pdata = pickle.load(fp)
        except Exception:
            return
        for item in pdata:
            if hasattr(self, item):
                setattr(self, item, pdata[item])


class PlaylistManager:
    """
    Manages saving and loading of playlists
    """

    def __init__(self, playlist_dir='playlists', playlist_class=Playlist):
        """
        Initializes the playlist manager

        @param playlist_dir: the data dir to save playlists to
        @param playlist_class: the playlist class to use
        """
        self.playlist_class = playlist_class
        self.playlist_dir = os.path.join(xdg.get_data_dirs()[0], playlist_dir)
        if not os.path.exists(self.playlist_dir):
            os.makedirs(self.playlist_dir)
        self.order_file = os.path.join(self.playlist_dir, 'order_file')
        self.playlists = []
        self.load_names()

    def _create_playlist(self, name):
        return self.playlist_class(name=name)

    def has_playlist_name(self, playlist_name):
        """
        Returns true if the manager has a playlist with the same name
        """
        return playlist_name in self.playlists

    def save_playlist(self, pl, overwrite=False):
        """
        Saves a playlist

        @param pl: the playlist
        @param overwrite: Set to [True] if you wish to overwrite a
            playlist should it happen to already exist
        """
        name = pl.name
        if overwrite or name not in self.playlists:
            pl.save_to_location(os.path.join(self.playlist_dir, encode_filename(name)))

            if name not in self.playlists:
                self.playlists.append(name)
            # self.playlists.sort()
            self.save_order()
        else:
            raise PlaylistExists

        event.log_event('playlist_added', self, name)

    def remove_playlist(self, name):
        """
        Removes a playlist from the manager, also
        physically deletes its

        @param name: the name of the playlist to remove
        """
        if name in self.playlists:
            try:
                os.remove(os.path.join(self.playlist_dir, encode_filename(name)))
            except OSError:
                pass
            self.playlists.remove(name)
            event.log_event('playlist_removed', self, name)

    def rename_playlist(self, playlist, new_name):
        """
        Renames the playlist to new_name
        """
        old_name = playlist.name
        if old_name in self.playlists:
            self.remove_playlist(old_name)
            playlist.name = new_name
            self.save_playlist(playlist)

    def load_names(self):
        """
        Loads the names of the playlists from the order file
        """
        # collect the names of all playlists in playlist_dir
        existing = []
        for f in os.listdir(self.playlist_dir):
            # everything except the order file should be a playlist, but
            # check against hidden files since some editors put
            # temporary stuff in the same dir.
            if f != os.path.basename(self.order_file) and not f.startswith("."):
                pl = self._create_playlist(f)
                path = os.path.join(self.playlist_dir, f)
                try:
                    pl.load_from_location(path)
                except Exception:
                    logger.exception("Failed loading playlist: %r", path)
                else:
                    existing.append(pl.name)

        # if order_file exists then use it
        if os.path.isfile(self.order_file):
            ordered_playlists = self.load_from_location(self.order_file)
            self.playlists = [n for n in ordered_playlists if n in existing]
        else:
            self.playlists = existing

    def get_playlist(self, name):
        """
        Gets a playlist by name

        @param name: the name of the playlist you wish to retrieve
        """
        if name in self.playlists:
            pl = self._create_playlist(name)
            pl.load_from_location(
                os.path.join(self.playlist_dir, encode_filename(name))
            )
            return pl
        else:
            raise ValueError("No such playlist '%s'" % name)

    def list_playlists(self):
        """
        Returns all the contained playlist names
        """
        return self.playlists[:]

    def move(self, playlist, position, after=True):
        """
        Moves the playlist to where position is
        """
        # Remove the playlist first
        playlist_index = self.playlists.index(playlist)
        self.playlists.pop(playlist_index)
        # insert it now after position
        position_index = self.playlists.index(position)
        if after:
            position_index = position_index + 1
        self.playlists.insert(position_index, playlist)

    def save_order(self):
        """
        Saves the order to the order file
        """
        self.save_to_location(self.order_file)

    def save_to_location(self, location):
        """
        Saves the names of the playlist to a file that is
        used to restore their order
        """
        if os.path.exists(location):
            f = open(location + ".new", "w")
        else:
            f = open(location, "w")
        for playlist in self.playlists:
            f.write(playlist)
            f.write('\n')

        f.write("EOF\n")
        f.close()
        if os.path.exists(location + ".new"):
            os.remove(location)
            os.rename(location + ".new", location)

    def load_from_location(self, location):
        """
        Loads the names of the playlist from a file.
        Their load order is their view order

        @return: a list of the playlist names
        """
        f = None
        for loc in [location, location + ".new"]:
            try:
                f = open(loc, 'r')
                break
            except Exception:
                pass
        if f is None:
            return []
        playlists = []
        while True:
            line = f.readline()
            if line == "EOF\n" or line == "":
                break
            playlists.append(line[:-1])
        f.close()
        return playlists


class SmartPlaylistManager(PlaylistManager):
    """
    Manages saving and loading of smart playlists
    """

    def __init__(self, playlist_dir, playlist_class=SmartPlaylist, collection=None):
        """
        Initializes a smart playlist manager

        @param playlist_dir: the data dir to save playlists to
        @param playlist_class: the playlist class to use
        @param collection: the default collection to use for searching
        """
        self.collection = collection
        PlaylistManager.__init__(
            self, playlist_dir=playlist_dir, playlist_class=playlist_class
        )

    def _create_playlist(self, name):
        # set a default collection so that get_playlist() always works
        return self.playlist_class(name=name, collection=self.collection)


# vim: et sts=4 sw=4



from gi.repository import GLib
from gi.repository import Gtk

import http.client
import logging
import operator
import os
import re
import socket
import urllib.parse
from xml.dom import minidom

from xl import common, event, playlist, xdg, trax
from xl.radio import RadioStation, RadioList, RadioItem
from xl.nls import gettext as _
from xlgui.widgets import dialogs

logger = logging.getLogger(__name__)
STATION = None


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, "exaile_loaded")
    else:
        _enable(None, exaile, None)


def _enable(devicename, exaile, nothing):
    global STATION

    STATION = IcecastRadioStation(exaile)
    exaile.radio.add_station(STATION)


def disable(exaile):
    global STATION
    exaile.radio.remove_station(STATION)
    STATION = None


def set_status(message, timeout=0):
    from xlgui.panel import radio

    radio.set_status(message, timeout)


class IcecastRadioStation(RadioStation):
    """
    Icecast Radio

    Simple usage:

    >>> manager = RadioManager()
    >>> manager.add_station(IcecastRadioStation())
    >>> stations = manager.search('icecast', 'ravetrax')
    >>> tracks = stations[0].get_playlist().get_tracks()
    >>> len(tracks) > 0
    True
    >>>
    """

    name = 'icecast'

    def __init__(self, exaile):
        """
        Initializes the icecast radio station
        """
        self.exaile = exaile
        self.user_agent = exaile.get_user_agent_string('icecast')
        self.icecast_url = 'https://dir.xiph.org'

        self.xml_url = self.icecast_url + '/yp.xml'
        self.genres_url = self.icecast_url + '/genres'

        self.cache_file = os.path.join(xdg.get_cache_dir(), 'icecast.cache')
        self.data = {}
        self._load_cache()
        self.playlists = {}

        logger.debug(self.user_agent)

    def _load_cache(self):
        """
        Loads icecast data from cache
        """
        if os.path.isfile(self.cache_file):
            try:
                dom = minidom.parse(self.cache_file)
            except:
                return
            self.data = {}
            for genre in dom.getElementsByTagName('genre'):
                genre_urls = {}

                for station in genre.getElementsByTagName('station'):
                    entry = {}
                    entry['url'] = station.getAttribute('url')
                    entry['bitrate'] = station.getAttribute('bitrate')
                    entry['format'] = station.getAttribute('format')

                    genre_urls[station.getAttribute('name')] = entry
                self.data[genre.getAttribute('name')] = genre_urls

    def _save_cache(self):
        """
        Saves cache data
        """
        impl = minidom.getDOMImplementation()
        document = impl.createDocument(None, 'genrelist', None)
        genrelist = document.documentElement
        for genre_name, stations in self.data.items():
            genre = document.createElement('genre')
            genre.setAttribute('name', genre_name)
            for station_name, url in stations.items():
                station = document.createElement('station')
                station.setAttribute('name', station_name)
                station.setAttribute('url', url['url'])
                station.setAttribute('bitrate', url['bitrate'])
                station.setAttribute('format', url['format'])
                genre.appendChild(station)
            genrelist.appendChild(genre)
        with open(self.cache_file, 'w', encoding="utf-8") as h:
            document.writexml(h, indent='\n', encoding='utf-8')

    def get_lists(self, no_cache=False):
        """
        Returns the stations list for icecast
        """
        # Level 1
        from xlgui.panel import radio

        if no_cache or not self.data:
            set_status(_('Contacting Icecast server...'))
            genre_list = self._get_genres()
            hostinfo = urllib.parse.urlparse(self.xml_url)

            c = http.client.HTTPConnection(hostinfo.netloc, timeout=20)
            try:
                c.request('GET', hostinfo.path, headers={'User-Agent': self.user_agent})
                response = c.getresponse()
            except (socket.timeout, socket.error):
                raise radio.RadioException(_('Error connecting to Icecast server.'))

            if response.status != 200:
                raise radio.RadioException(_('Error connecting to Icecast server.'))

            set_status(_('Parsing stations XML...'))
            body = response.read()
            c.close()

            def get_text(node):
                for subnode in node.childNodes:
                    if subnode.nodeType == minidom.Node.TEXT_NODE:
                        return subnode.nodeValue

            data = {}
            dom = minidom.parseString(body)
            entries = dom.getElementsByTagName('entry')
            set_status(_('Retrieving stations...'))
            for entry in entries:
                url_node = entry.getElementsByTagName('listen_url')[0]
                name_node = entry.getElementsByTagName('server_name')[0]
                genre_node = entry.getElementsByTagName('genre')[0]
                bitrate_node = entry.getElementsByTagName('bitrate')[0]
                server_type_node = entry.getElementsByTagName('server_type')[0]

                name = get_text(name_node)
                url = get_text(url_node)
                genre = get_text(genre_node) or _('Unknown')
                bitrate = get_text(bitrate_node)
                format = get_text(server_type_node)

                station_genres = genre.split(' ')
                insert_genres = []
                for i in range(0, len(station_genres)):
                    if station_genres[i].lower() in genre_list:
                        insert_genres.append(station_genres[i])
                    if i < len(station_genres) - 1:
                        if (
                            (station_genres[i] + " " + station_genres[i + 1]).lower()
                        ) in genre_list:
                            insert_genres.append(
                                station_genres[i] + " " + station_genres[i + 1]
                            )

                entry = {}
                entry['url'] = url
                entry['bitrate'] = bitrate
                entry['format'] = format
                for genre in insert_genres:
                    if not genre in data:
                        data[genre] = {}
                    data[genre][name] = entry

            self.data = data
            self._save_cache()
        else:
            data = self.data
        rlists = []

        for item in data.keys():
            if item is None:
                continue
            rlist = RadioList(item, station=self)
            rlist.get_items = lambda no_cache, name=item: self._get_subrlists(
                name=name, no_cache=no_cache
            )
            rlists.append(rlist)

        rlists.sort(key=operator.attrgetter('name'))
        self.rlists = rlists
        set_status('')
        return rlists

    def _get_subrlists(self, name, no_cache=False):
        """
        Gets the stations to a genre
        """
        sublist = self.data[name]
        station_list = self._build_station_list(sublist)
        return station_list

    def _get_playlist(self, name, station_url):
        """
        Gets the playlist for the given name and id
        """
        if station_url in self.playlists:
            return self.playlists[station_url]
        set_status(_('Contacting Icecast server...'))

        track = trax.Track(station_url)
        pls = playlist.Playlist(name=name)
        pls.append(track)
        self.playlists[station_url] = pls

        set_status('')
        return pls

    def search(self, keyword):
        """
        Searches the station for a specified keyword

        @param keyword: the keyword to search
        """
        items = {}
        for genre, stations in self.data.items():
            for name, station in stations.items():
                if keyword.lower() in name.lower():
                    items[name] = station

        return self._build_station_list(items)

    def _build_station_list(self, list):
        station_list = []
        for station_name, url in list.items():
            stat = RadioItem(station_name, url['url'])
            stat.bitrate = url['bitrate']
            stat.format = url['format']
            stat.get_playlist = lambda name=station_name, station_url=url[
                'url'
            ]: self._get_playlist(name, station_url)

            station_list.append(stat)
        station_list.sort(key=operator.attrgetter('name'))
        return station_list

    def on_search(self):
        """
        Called when the user wants to search for a specific stream
        """
        dialog = dialogs.TextEntryDialog(
            _("Enter the search keywords"), _("Icecast Search")
        )

        result = dialog.run()
        if result == Gtk.ResponseType.OK:
            keyword = dialog.get_value()

            self.do_search(keyword)

    @common.threaded
    def do_search(self, keyword):
        """
        Actually performs the search in a separate thread
        """
        lists = self.search(keyword)

        GLib.idle_add(self.search_done, keyword, lists)

    @common.idle_add()
    def search_done(self, keyword, lists):
        """
        Called when the search is finished
        """
        if not lists:
            dialogs.info(self.exaile.gui.main.window, _('No Stations Found'))
            return

        dialog = ResultsDialog(_("Icecast Search Results"))
        dialog.set_items(lists)
        dialog.connect('response', self._search_response)
        dialog.show_all()
        self._keyword = keyword

    def _search_response(self, dialog, result, *e):
        dialog.hide()
        if result == Gtk.ResponseType.OK:
            items = dialog.get_items()
            if not items:
                return

            self.do_get_playlist(self._keyword, items[0])

    @common.threaded
    def do_get_playlist(self, keyword, item):
        pl = item.get_playlist()
        if not pl:
            return

        GLib.idle_add(self.done_getting_playlist, pl)

    @common.idle_add()
    def done_getting_playlist(self, pl):
        self._parent.emit('playlist-selected', pl)

    def get_menu(self, parent):
        """
        Returns a menu that works for icecast radio
        """
        self._parent = parent
        menu = parent.get_menu()
        menu.add_simple(_("Search"), lambda *e: self.on_search(), Gtk.STOCK_FIND)
        return menu

    def _get_genres(self):
        from xlgui.panel import radio

        set_status(_('Contacting Icecast server...'))
        hostinfo = urllib.parse.urlparse(self.genres_url)

        c = http.client.HTTPConnection(hostinfo.netloc, timeout=20)
        try:
            c.request('GET', hostinfo.path, headers={'User-Agent': self.user_agent})
            response = c.getresponse()
        except (socket.timeout, socket.error):
            raise radio.RadioException(_('Error connecting to Icecast server.'))

        if response.status != 200:
            raise radio.RadioException(_('Error connecting to Icecast server.'))

        set_status(_('Parsing genre list...'))
        body = response.read()
        c.close()

        genres = {}
        next = False
        for line in body.splitlines():
            if next:
                genre = line.strip().decode('UTF-8')
                genres[genre.lower()] = genre
                next = False
            if b'list-group-item list-group-item-action' in line:
                next = True

        return genres


class ResultsDialog(dialogs.ListDialog):
    def __init__(self, title):
        dialogs.ListDialog.__init__(self, title)
        col = self.list.get_column(0)
        col.set_title(_('Name'))
        col.set_expand(True)
        col.set_resizable(True)
        self.list.set_headers_visible(True)
        text = Gtk.CellRendererText()
        text.set_property('xalign', 1.0)
        col = Gtk.TreeViewColumn(_('Bitrate'), text)
        col.set_cell_data_func(
            text,
            lambda column, cell, model, iter, unused: cell.set_property(
                'text', model.get_value(iter, 0).bitrate
            ),
        )
        self.list.append_column(col)
        text = Gtk.CellRendererText()
        text.set_property('xalign', 0.5)
        col = Gtk.TreeViewColumn(_('Format'), text)
        col.set_cell_data_func(
            text,
            lambda column, cell, model, iter, unused: cell.set_property(
                'text', model.get_value(iter, 0).format
            ),
        )
        self.list.append_column(col)



# Copyright (C) 2011 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk

import os
import os.path
import datetime

from xl import event, player, providers, settings, xdg

from xl.playlist import Playlist
from xl.nls import gettext as _

import xlgui
from xlgui import main
from xlgui.widgets import menu
from xlgui.widgets.notebook import NotebookTab
from xlgui.widgets.playlist import PlaylistPageBase, PlaylistView

from . import history_preferences


class HistoryPlugin:
    '''Implements logic for plugin'''

    def get_preferences_pane(self):
        return history_preferences

    def enable(self, exaile):
        self.exaile = exaile

    def on_gui_loaded(self):
        save_on_exit = settings.get_option(
            'plugin/history/save_on_exit', history_preferences.save_on_exit_default
        )
        shown = settings.get_option('plugin/history/shown', False)

        # immutable playlist that stores everything we've played
        self.history_loc = os.path.join(xdg.get_data_dir(), 'history')

        self.history_playlist = HistoryPlaylist(player.PLAYER)

        if save_on_exit:
            self.history_playlist.load_from_location(self.history_loc)

        self.history_tab = None
        self.history_page = None

        # add menu item to 'view' to display our playlist
        self.menu = menu.check_menu_item(
            'history',
            '',
            _('Playback history'),
            lambda *e: self.is_shown(),
            self.on_playback_history,
        )

        providers.register('menubar-view-menu', self.menu)

    def on_exaile_loaded(self):
        # add the history playlist to the primary notebook
        save_on_exit = settings.get_option(
            'plugin/history/save_on_exit', history_preferences.save_on_exit_default
        )
        shown = settings.get_option('plugin/history/shown', False)

        if save_on_exit and shown:
            self.show_history(True)

    def teardown(self, exaile):
        '''Called when exaile is exiting'''

        if settings.get_option(
            'plugin/history/save_on_exit', history_preferences.save_on_exit_default
        ):
            self.history_playlist.save_to_location(self.history_loc)
            settings.set_option('plugin/history/shown', self.is_shown())
        else:
            settings.set_option('plugin/history/shown', False)

        self.show_history(False)

    def disable(self, exaile):
        '''Called when the plugin is disabled'''
        if self.menu:
            providers.unregister('menubar-view-menu', self.menu)
            self.menu = None

        self.show_history(False)

        if os.path.exists(self.history_loc):
            # TODO: The parent should be the Preferences window, but we don't
            # have access to it, so this just uses the main window.
            dialog = Gtk.MessageDialog(
                buttons=Gtk.ButtonsType.YES_NO,
                message_type=Gtk.MessageType.QUESTION,
                modal=True,
                text=_('Erase stored history?'),
                transient_for=exaile.gui.main.window,
            )

            if dialog.run() == Gtk.ResponseType.YES:
                try:
                    os.unlink(self.history_loc)
                except Exception:
                    pass

            dialog.destroy()

    def is_shown(self):
        if self.history_page is None:
            return False
        return main.get_playlist_notebook().page_num(self.history_page) != -1

    def on_playback_history(self, menu, name, parent, context):
        self.show_history(not self.is_shown())

    def show_history(self, show):
        if show == self.is_shown():
            return

        pn = main.get_playlist_notebook()
        if show:
            self.history_page = HistoryPlaylistPage(
                self.history_playlist, player.PLAYER
            )
            self.history_tab = NotebookTab(
                main.get_playlist_notebook(), self.history_page
            )
            pn.add_tab(self.history_tab, self.history_page)
        else:
            pn.remove_tab(self.history_tab)


plugin_class = HistoryPlugin


class HistoryPlaylistPage(PlaylistPageBase):
    # add two buttons on the bottom: 'save to playlist', and
    # clear history. Use the dirty key to figure out if we
    # warn about clearing history..

    reorderable = False

    def __init__(self, playlist, player):
        PlaylistPageBase.__init__(self, playlist, player)

        self.playlist = playlist

        self.swindow = Gtk.ScrolledWindow()
        self.swindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.pack_start(self.swindow, True, True, 0)

        self.view = PlaylistView(self.playlist, player)
        self.swindow.add(self.view)

        hbox = Gtk.ButtonBox()
        hbox.set_halign(Gtk.Align.END)

        button = Gtk.Button(stock=Gtk.STOCK_CLEAR)
        button.connect('clicked', self.on_clear_history)
        hbox.pack_start(button, True, True, 0)

        button = Gtk.Button(stock=Gtk.STOCK_SAVE)
        button.connect('clicked', self.on_save_history)
        hbox.pack_start(button, True, True, 0)

        self.pack_start(hbox, False, False, 0)

        self.show_all()

    ## NotebookPage API ##

    def get_page_name(self):
        return _("History")

    ## End NotebookPage ##

    def on_save(self):
        self.save_history()

    def on_clear_history(self, widget):
        self.playlist.clear()

    def on_save_history(self, widget):
        self.save_history()

    def save_history(self):
        if len(self.playlist) == 0:
            return

        name = 'History %s' % datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

        playlists = xlgui.get_controller().get_panel('playlists')
        playlists.add_new_playlist(self.playlist, name)


def __create_history_tab_context_menu():
    smi = menu.simple_menu_item
    items = []
    items.append(
        smi(
            'save',
            [],
            _("Save History"),
            'gtk-save',
            lambda w, n, o, c: o.save_history(),
        )
    )
    items.append(
        smi(
            'clear',
            ['save'],
            _("_Clear History"),
            'gtk-clear',
            lambda w, n, o, c: o.playlist._clear(),
        )
    )


class HistoryPlaylist(Playlist):
    def __init__(self, player):
        Playlist.__init__(self, _('History'))

        # catch the history
        event.add_ui_callback(
            self.__on_playback_track_start, 'playback_track_start', player
        )

        if player.is_paused() or player.is_playing():
            self.__on_playback_track_start(
                'playback_track_start', player, player.current
            )

    def __on_playback_track_start(self, event, player, track):
        '''Every time a track plays, add it to the playlist'''

        maxlen = int(
            settings.get_option(
                'plugin/history/history_length',
                history_preferences.history_length_default,
            )
        )
        if maxlen < 0:
            maxlen = 0
            settings.set_option('plugin/history/history_length', 0)

        if len(self) >= maxlen - 1:
            Playlist.__delitem__(self, slice(0, max(0, len(self) - (maxlen - 1)), None))

        Playlist.__setitem__(self, slice(len(self), len(self), None), [track])

    def clear(self):
        Playlist.__delitem__(self, slice(None, None, None))

    #
    # Suppress undesirable playlist functions, history playlist is immutable
    #

    def set_shuffle_mode(self, mode):
        pass

    def set_repeat_mode(self, mode):
        pass

    def set_dynamic_mode(self, mode):
        pass

    def randomize(self, positions=None):
        pass

    def sort(self, tags, reverse=False):
        pass

    def __setitem__(self, i, value):
        pass

    def __delitem__(self, i):
        pass

    def append(self, other):
        pass

    def extend(self, other):
        pass

    def pop(self, i=-1):
        pass



# karaoke - Removes voice from Exaile's audio output
# Copyright (C) 2009-2010  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from gi.repository import Gst

import xl.providers
from xl.player.gst.gst_utils import ElementBin


class Karaoke(ElementBin):
    index = 50
    name = 'karaoke'

    def __init__(self):
        ElementBin.__init__(self, name=self.name)
        self.elements[50] = Gst.ElementFactory.make('audiokaraoke', None)
        self.setup_elements()


def enable(exaile):
    xl.providers.register('gst_audio_filter', Karaoke)


def disable(exaile):
    xl.providers.unregister('gst_audio_filter', Karaoke)


# vi: et sts=4 sw=4 tw=80



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from typing import Literal

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from xl import event, formatter, main, settings, xdg
from xl.nls import gettext as _, ngettext
import xlgui
from xlgui import cover, guiutil
from xlgui.widgets import playlist, queue
from xlgui.widgets.playback import PlaybackProgressBar


class TrackInfoPane(Gtk.Bin):
    """
    Displays cover art and track data
    """

    def __init__(self, player):
        Gtk.Bin.__init__(self)
        self.__player = player

        builder = guiutil.get_builder(
            xdg.get_data_path('ui', 'widgets', 'track_info.ui')
        )
        info_box = builder.get_object('info_box')
        info_box.get_parent().remove(info_box)
        self.add(info_box)

        self.__auto_update = False
        self.__display_progress = False
        self.__formatter = formatter.TrackFormatter(
            _(
                '<span size="x-large" weight="bold">$title</span>\n'
                'by $artist\n'
                'from $album'
            )
        )
        self.__formatter.connect('notify::format', self.on_notify_format)
        self.__default_text = '<span size="x-large" ' 'weight="bold">%s</span>\n\n' % _(
            'Not Playing'
        )
        self.__cover_size = None
        self.__timer = None
        self.__track = None

        self.info_label = builder.get_object('info_label')
        self.action_area = builder.get_object('action_area')
        self.progress_box = builder.get_object('progress_box')
        self.playback_image = builder.get_object('playback_image')
        self.progressbar = PlaybackProgressBar(player)
        guiutil.gtk_widget_replace(builder.get_object('progressbar'), self.progressbar)

        self.cover = cover.CoverWidget(builder.get_object('cover_image'))
        self.cover.hide()
        self.cover.set_no_show_all(True)

        self.clear()
        self.__update_widget_state()

    def destroy(self):
        """
        Cleanups
        """
        # Make sure to disconnect callbacks
        self.set_auto_update(False)

        Gtk.Bin.destroy(self)

    def get_auto_update(self):
        """
        Gets whether the info pane shall
        be automatically updated or not

        :rtype: bool
        """
        return self.__auto_update

    def set_auto_update(self, auto_update):
        """
        Sets whether the info pane shall
        be automatically updated or not

        :param auto_update: enable or disable
            automatic updating
        :type auto_update: bool
        """
        if auto_update != self.__auto_update:
            self.__auto_update = auto_update

            p_evts = [
                'playback_player_end',
                'playback_track_start',
                'playback_toggle_pause',
                'playback_error',
            ]
            events = ['track_tags_changed', 'cover_set', 'cover_removed']

            if auto_update:
                for e in p_evts:
                    event.add_ui_callback(getattr(self, 'on_%s' % e), e, self.__player)
                for e in events:
                    event.add_ui_callback(getattr(self, 'on_%s' % e), e)

                self.set_track(self.__player.current)
            else:
                for e in p_evts:
                    event.remove_callback(getattr(self, 'on_%s' % e), e, self.__player)
                for e in events:
                    event.remove_callback(getattr(self, 'on_%s' % e), e)

    def get_cover_size(self):
        """
        Gets the preferred cover size

        :rtype: int
        """
        return self.__cover_size or settings.get_option('gui/cover_width', 100)

    def set_cover_size(self, cover_size):
        """
        Overrides the cover size to display,
        set to None to use global default

        :param cover_size: the preferred cover size
        :type cover_size: int
        """
        self.__cover_size = cover_size

    def get_default_text(self):
        """
        Gets the default text displayed
        when the playback is stopped

        :rtype: string
        """
        return self.__default_text

    def set_default_text(self, default_text):
        """
        Sets the default text displayed
        when the playback is stopped

        :param default_text: the new default text
        :type default_text: string
        """
        self.__default_text = default_text

    def get_display_progress(self):
        """
        Returns whether the progress indicator
        is currently visible or not

        :rtype: bool
        """
        return self.__display_progress

    def set_display_progress(self, display_progress):
        """
        Shows or hides the progress indicator. The
        indicator will not be displayed if the
        currently displayed track is not playing.

        :param display_progress: Whether to show
            or hide the progress indicator
        :type display_progress: bool
        """
        self.__display_progress = display_progress
        self.__update_widget_state()

    def get_info_format(self):
        """
        Gets the current format used
        to display the track data

        :rtype: string
        """
        return self.__formatter.props.format

    def set_info_format(self, format):
        """
        Sets the format used to display the track data

        :param format: the format, see the documentation
            of :class:`string.Template` for details
        :type format: string
        """
        self.__formatter.props.format = format

    def set_track(self, track):
        """
        Updates the data displayed in the info pane

        :param track: A track to take the data from,
            clears the info pane if track is None
        :type track: :class:`xl.trax.Track`
        """
        if track is None:
            self.clear()
            return

        self.__track = track

        self.cover.set_track(track)

        self.info_label.set_markup(self.__formatter.format(track, markup_escape=True))
        self.__update_widget_state()

    def __update_widget_state(self):
        if self.__display_progress:
            if self.__track == self.__player.current and not self.__player.is_stopped():
                if self.__player.is_paused():
                    icon_name = 'media-playback-pause'
                else:
                    icon_name = 'media-playback-start'
                self.playback_image.set_from_icon_name(
                    icon_name, Gtk.IconSize.SMALL_TOOLBAR
                )

            self.progress_box.set_no_show_all(False)
            self.progress_box.set_visible(True)
            self.progressbar.set_visible(True)
        else:
            self.progress_box.set_visible(False)
            self.progress_box.set_no_show_all(True)
            self.progressbar.set_visible(True)

    def clear(self):
        """
        Resets the info pane
        """

        self.cover.set_track(None)
        self.info_label.set_markup(self.__default_text)

        self.__track = None
        self.__update_widget_state()

    def get_action_area(self):
        """
        Retrieves the action area
        at the end of the pane

        :rtype: :class:`Gtk.Box`
        """
        return self.action_area

    def on_notify_format(self, formatter, format):
        """
        Updates the displayed data after format changes
        """
        if self.__track is not None:
            self.set_track(self.__track)

    def on_playback_player_end(self, event, player, track):
        """
        Clears the info pane on playback end
        """
        self.clear()

    def on_playback_track_start(self, event, player, track):
        """
        Updates the info pane on track start
        """
        self.set_track(track)

    def on_playback_toggle_pause(self, event, player, track):
        """
        Updates the info pane on playback pause/resume
        """
        self.set_track(track)

    def on_playback_error(self, event, player, message):
        """
        Clears the info pane on playback errors
        """
        self.clear()

    def on_track_tags_changed(self, event, track, tags):
        """
        Updates the info pane on tag changes
        """
        if (
            self.__player is not None
            and not self.__player.is_stopped()
            and track is self.__track
        ):
            self.set_track(track)

    def on_cover_set(self, event, covers, track):
        """
        Updates the info pane on cover set
        """
        if track is self.__track:
            self.set_track(track)

    def on_cover_removed(self, event, covers, track):
        """
        Updates the info pane on cover removal
        """
        if track is self.__track:
            self.set_track(track)


class ToolTip:
    """
    Custom tooltip class to allow for
    extended tooltip functionality
    """

    def __init__(self, parent, widget):
        """
        :param parent: the parent widget the tooltip
            should be attached to
        :param widget: the tooltip widget to be used
            for the tooltip
        """
        if self.__class__.__name__ == 'ToolTip':
            raise TypeError(
                "cannot create instance of abstract "
                "(non-instantiable) type `ToolTip'"
            )

        self.__widget = widget
        self.__widget.unparent()  # Just to be sure

        parent.set_has_tooltip(True)
        parent.connect('query-tooltip', self.on_query_tooltip)

    def on_query_tooltip(self, widget, x, y, keyboard_mode, tooltip):
        """
        Puts the custom widget into the tooltip
        """
        tooltip.set_custom(self.__widget)

        return True


class TrackToolTip(TrackInfoPane, ToolTip):
    """
    Track specific tooltip class, displays
    track data and progress indicators
    """

    def __init__(self, parent, player):
        """
        :param parent: the parent widget the tooltip
            should be attached to
        """
        TrackInfoPane.__init__(self, player)
        ToolTip.__init__(self, parent, self)

        self.set_border_width(6)
        self.info_label.set_ellipsize(Pango.EllipsizeMode.NONE)
        self.cover.set_no_show_all(False)
        self.cover.show_all()

    def destroy(self):
        """
        Cleanups
        """
        TrackInfoPane.destroy(self)


class StatusbarTextFormatter(formatter.Formatter):
    """
    A text formatter for status indicators
    """

    def __init__(self, format):
        """
        :param format: The initial format, see the documentation
            of string.Template for details
        :type format: string
        """
        formatter.Formatter.__init__(self, format)

        self._substitutions = {
            'collection_count': self.get_collection_count,
            'playlist_count': self.get_playlist_count,
            'playlist_duration': self.get_playlist_duration,
        }

    def get_collection_count(self):
        """
        Retrieves the collection count
        """
        if not settings.get_option('gui/show_status_bar_collection_count', True):
            return ''
        return _('%d in collection') % main.exaile().collection.get_count()

    def get_playlist_count(
        self, selection: Literal['none', 'override', 'only'] = 'none'
    ):
        """
        Retrieves the count of tracks in either the
        full playlist or the current selection

        :param selection: 'none' for playlist count only,
            'override' for selection count if tracks are selected,
            playlist count otherwise, 'only' for selection count only
        """
        if not settings.get_option(
            'gui/show_status_bar_count_tracks_in_playlist', True
        ):
            return ""

        page = xlgui.main.get_selected_page()

        if not isinstance(page, playlist.PlaylistPageBase):
            return ""

        playlist_count = len(page.playlist)
        selection_count = page.view.get_selection_count()

        if selection == 'none':
            return ngettext("%d showing", "%d showing", playlist_count) % playlist_count
        elif selection == 'override':
            if selection_count > 1:
                return (
                    ngettext("%d selected", "%d selected", selection_count)
                    % selection_count
                )
            else:
                return (
                    ngettext("%d showing", "%d showing", playlist_count)
                    % playlist_count
                )
        elif selection == 'only':
            return (
                ngettext("%d selected", "%d selected", selection_count)
                % selection_count
            )

        raise ValueError(
            'Invalid argument "%s" passed to parameter '
            '"selection" for "playlist_count", possible arguments are '
            '"none", "override" and "only"' % selection
        )

    def get_playlist_duration(self, format='short', selection='none'):
        """
        Retrieves the duration of all tracks in
        the playlist or within the selection

        :param format: Verbosity of the output
            Possible values are short, long or verbose
            yielding "1:02:42", "1h, 2m, 42s" or
            "1 hour, 2 minutes, 42 seconds"
        :type format: string
        :param selection: 'none' for playlist count only,
            'override' for selection count if tracks are selected,
            playlist count otherwise, 'only' for selection count only
        :type selection: string
        """
        if not settings.get_option('gui/show_status_bar_time_in_playlist', True):
            return ''

        page = xlgui.main.get_selected_page()

        if not isinstance(page, playlist.PlaylistPageBase):
            return ''

        playlist_duration = sum(t.get_tag_raw('__length') or 0 for t in page.playlist)
        selection_tracks = page.view.get_selected_tracks()
        selection_count = len(selection_tracks)
        selection_duration = sum(
            t.get_tag_raw('__length') or 0 for t in selection_tracks
        )

        if selection == 'none':
            duration = playlist_duration
        elif selection == 'override':
            if selection_count > 1:
                duration = selection_duration
            else:
                duration = playlist_duration
        elif selection == 'only':
            if selection_count > 1:
                duration = selection_duration
            else:
                duration = 0
        else:
            raise ValueError(
                'Invalid argument "%s" passed to parameter '
                '"selection" for "playlist_duration", possible arguments are '
                '"none", "override" and "only"' % selection
            )

        return formatter.LengthTagFormatter.format_value(duration, format)


class Statusbar:
    """
    Convenient access to multiple status labels
    """

    def __init__(self, status_bar):
        """
        Initialises the status bar
        """
        self.status_bar = status_bar

        self.formatter = StatusbarTextFormatter(self._get_substitutions())

        self.info_label = Gtk.Label()

        box = self.status_bar.get_message_area()

        box.pack_end(self.info_label, False, True, 0)

        self.context_id = self.status_bar.get_context_id('status')
        self.message_ids = []

        event.add_callback(self._on_option_set, "gui_option_set")

    def _get_substitutions(self) -> str:
        sub = settings.get_option('gui/statusbar_info_format', '')

        if sub:
            return sub

        show_playlist_count = settings.get_option(
            'gui/show_status_bar_count_tracks_in_playlist', True
        )
        show_playlist_duration = settings.get_option(
            'gui/show_status_bar_time_in_playlist', True
        )
        show_collection_count = settings.get_option(
            'gui/show_status_bar_collection_count', True
        )

        if show_playlist_count:
            sub = sub + '${playlist_count:selection=override}'

        if show_playlist_count and show_playlist_duration:
            sub = (
                sub
                + ' ${playlist_duration:selection=override, format=long, prefix=(, suffix=), pad=0 }'
            )

        elif not show_playlist_count and show_playlist_duration:
            sub = sub + '${playlist_duration:selection=override, format=long, pad=0 }'

        if show_collection_count and (show_playlist_count or show_playlist_duration):
            sub = sub.rstrip() + ', '

        if show_collection_count:
            sub = sub + '$collection_count'

        return sub

    def _on_option_set(self, name, object, option: str):
        if not option in [
            'gui/show_status_bar_collection_count',
            'gui/show_status_bar_count_tracks_in_playlist',
            'gui/show_status_bar_time_in_playlist',
        ]:
            return

        self.formatter = StatusbarTextFormatter(self._get_substitutions())

        self.update_info()

    def set_status(self, status, timeout=0):
        """
        Sets the status message
        """
        self.message_ids += [self.status_bar.push(self.context_id, status)]

        if timeout > 0:
            GLib.timeout_add_seconds(timeout, self.clear_status)

    def clear_status(self):
        """
        Clears the status message
        """
        try:
            for message_id in self.message_ids:
                self.status_bar.remove_message(self.context_id, message_id)
        except AttributeError:
            for message_id in self.message_ids:
                self.status_bar.remove(self.context_id, message_id)

        del self.message_ids[:]
        self.message_ids = []

    def update_info(self):
        """
        Updates the info label text
        """
        if settings.get_option('gui/show_status_bar', True):
            self.status_bar.show()
        else:
            self.status_bar.hide()

        self.info_label.set_label(self.formatter.format())

    def show(self):
        self.status_bar.show()

    def hide(self):
        self.status_bar.hide()


# TODO: Check if we can get a progress indicator in here somehow
class Splash:
    """
    A splash screen suitable for indicating startup;
    will automatically be destroyed after GUI startup
    """

    def __init__(self):
        builder = guiutil.get_builder(xdg.get_data_path('ui', 'splash.ui'))

        image = builder.get_object('splash_image')
        image.set_from_file(xdg.get_data_path('images', 'splash.png'))

        self.window = builder.get_object('SplashScreen')

    def destroy(self):
        """
        Destroys the splash screen
        """
        self.window.destroy()

    def show(self):
        """
        Shows the splash screen
        """
        # Show the splash screen without causing startup notification.
        Gtk.Window.set_auto_startup_notification(False)
        self.window.show_all()
        Gtk.Window.set_auto_startup_notification(True)

        # Ensure the splash is completely drawn before moving on
        while Gtk.events_pending():
            Gtk.main_iteration()

    def hide(self):
        """
        Hides the splash screen
        """
        self.window.hide()
